from scm.plams import Molecule, Settings, AMSNVTJob, AMSNVEJob, AMSAnalysisJob
from scm.plams import init, finish

text = """6

O          0.0000000000        0.0000000000        0.0000000000
H          0.5580000000        0.5980000000       -0.5130000000
H          0.0280000000       -0.8150000000       -0.5120000000
O         -0.0660000000        0.0230000000        2.7240000000
H          0.0000000000        0.0000000000        1.7490000000
H         -0.9560000000        0.3570000000        2.8380000000
"""


def main():
    """
    The main script
    """
    # Write the XYZ file
    xyzfile = open("2h2o.xyz", "w")
    xyzfile.write(text)
    xyzfile.close()

    # Initialize PLAMS
    init()

    # Set up and run the equilibration at 300K
    mol = Molecule("2h2o.xyz")
    settings = Settings()
    settings.input.DFTB.Model = "GFN1-xTB"
    job = AMSNVTJob(molecule=mol, settings=settings, nsteps=10000, timestep=0.5, thermostat="Berendsen")
    results = job.run()

    # Run the production run, without thermostat
    job = AMSNVEJob.restart_from(job, binlog_dipolemoment=True)
    results = job.run()

    # Set up and run the analysis
    ansett = Settings()
    ansett.input.TrajectoryInfo.Trajectory.KFFilename = results.rkfpath()
    ansett.input.Task = "AutoCorrelation"
    ansett.input.AutoCorrelation.Property = "DipoleMomentFromBinLog"
    ansett.input.AutoCorrelation.WritePropertyToKF = "Yes"
    ansett.input.AutoCorrelation.UseTimeDerivative.Enabled = "Yes"
    ansett.input.AutoCorrelation.WritePropertyToKF = "Yes"

    anjob = AMSAnalysisJob(settings=ansett)
    anresults = anjob.run()

    # Write the analysis plots
    plots = anresults.get_all_plots()
    for xy in plots:
        xy.write("%s" % (xy.section + ".txt"))

        # Finalize PLAMS
        finish()


if __name__ == "__main__":
    main()



#!/usr/bin/env amspython
from scm.plams import *

"""
    Script to get and print the first 5 frames as PLAMS Molecules
    from an ams.rkf trajectory.

    The script also shows how to print a speciic frame, frame 23.

    Set the correct path to ams.rkf below, and then run as
    $AMSBIN/amspython MoleculesFromRKFTrajectory.py
"""


def main():
    job = AMSJob.load_external("ams.rkf")  # modify to give the path to the ams.rkf file

    trajectory = Trajectory(job.results.rkfpath())
    for i, mol in enumerate(trajectory, 1):
        print(f"frame {i}")
        print(mol)  # mol is a PLAMS Molecule
        if i == 5:
            break

    print("Extracting specific molecule")
    mol = job.results.get_history_molecule(23)
    print("Frame 23")
    print(mol)


if __name__ == "__main__":
    main()



import os
import sys

from scm.plams import (
    AMSJob,
    CRSJob,
    KFFile,
    Molecule,
    Results,
    Settings,
    Units,
    config,
    finish,
    init,
)

__all__ = ["OxidationPotentialCalculator"]


class OxidationPotentialCalculator:
    def __init__(self, logfile=sys.stdout) -> None:

        self.logfile = logfile  # file to send prints to, default None disables logging (print still works)
        self.set_default_settings()  # after this function has been called it is possible to overwrite settings

    def __call__(self, *args, **kwargs) -> float:
        return self.oxidation_potential(*args, **kwargs)

    def set_default_settings(self) -> None:
        #### DEFAULT SETTINGS, these can be changed after creating a new OxidationPotentialCalculator object
        # default settings for pre-optimization using DFTB
        self.pre_optimize_defaults = Settings()
        self.pre_optimize_defaults.input.ams.task = "GeometryOptimization"
        self.pre_optimize_defaults.input.ams.Properties.NormalModes = "Yes"
        self.pre_optimize_defaults.input.ams.Properties.PESPointCharacter = "Yes"
        self.pre_optimize_defaults.input.ams.NormalModes.ReScanFreqRange = "-1000 0"
        self.pre_optimize_defaults.input.ams.PESPointCharacter.NegativeEigenvalueTolerance = -0.001
        self.pre_optimize_defaults.input.DFTB
        self.pre_optimize_defaults.input.DFTB.Model = "DFTB3"
        self.pre_optimize_defaults.input.DFTB.ResourcesDir = "DFTB.org/3ob-3-1"

        # DFTB GO settings
        self.DFTB_defaults = Settings()
        self.DFTB_defaults.input.ams.task = "GeometryOptimization"
        self.DFTB_defaults.input.DFTB
        self.DFTB_defaults.input.DFTB.Model = "GFN1-xTB"

        # default settings for optimization and singlepoint using DFT
        self.DFT_defaults = Settings()
        self.DFT_defaults.input.ams.task = "GeometryOptimization"
        self.DFT_defaults.input.adf.basis.type = "TZ2P"
        self.DFT_defaults.input.adf.basis.core = "None"
        # self.DFT_defaults.input.adf.SCF.converge     = '1e-7 1e-7'
        # self.DFT_defaults.input.adf.SCF.mixing       = '0.015'
        # self.DFT_defaults.input.adf.SCF.DIIS.N       = '25'
        self.DFT_defaults.input.adf.xc.hybrid = "B3LYP"
        self.DFT_defaults.input.adf.xc.Dispersion = "GRIMME3 BJDAMP"
        self.DFT_defaults.input.adf.Relativity.Level = "None"
        # self.DFT_defaults.input.ams.GeometryOptimization.Coordinatetype = 'Cartesian'
        self.DFT_defaults.input.adf.NumericalQuality = "Good"
        # self.DFT_defaults.input.adf.RIHartreeFock.UseMe      = 'Yes'
        # self.DFT_defaults.input.adf.RIHartreeFock.Quality    = 'Normal'
        self.DFT_defaults.input.adf.Symmetry = "NOSYM"
        self.DFT_defaults.input.ams.UseSymmetry = "No"

        # frequency calculation settings
        self.frequencies_defaults = Settings()
        self.frequencies_defaults.input.ams.properties.NormalModes = "Yes"
        self.frequencies_defaults.input.ams.Properties.PESPointCharacter = "No"
        self.frequencies_defaults.input.ams.NormalModes.ReScanFreqRange = "-1000 0"
        self.frequencies_defaults.input.ams.PESPointCharacter.NegativeEigenvalueTolerance = -0.001

        # default solvent settings for optimization
        self.COSMO_defaults = Settings()
        self.COSMO_defaults.input.adf.Solvation.Solv = "Name=Dichloromethane"

        # default settings for oxidized molecules
        self.DFT_oxidized_defaults = Settings()
        self.DFT_oxidized_defaults.input.adf.Unrestricted = "Yes"
        self.DFT_oxidized_defaults.input.adf.SpinPolarization = "1.0"
        self.DFT_oxidized_defaults.input.ams.System.Charge = "1.0"

        # default settings for DFTB for oxidized molecules
        self.DFTB_oxidized_defaults = Settings()
        self.DFTB_oxidized_defaults.input.ams.System.Charge = "1.0"

    def set_paths(self, job_dir: str = os.getcwd()) -> None:

        # default paths
        self.job_dir = job_dir
        if not os.path.exists(job_dir):
            os.makedirs(job_dir)
        self.log(f"Job directory:                             {job_dir}")

        self.geo_preopt_dir = os.path.join(job_dir, "Geometries", "Preoptimization", "Preopt_Geo")
        if not os.path.exists(self.geo_preopt_dir):
            os.makedirs(self.geo_preopt_dir)
        self.log(f"Preoptimization geometry directory:        {self.geo_preopt_dir}")

        self.not_geo_preopt_dir = os.path.join(job_dir, "Geometries", "Preoptimization", "Not_Preopt_Geo")
        if not os.path.exists(self.not_geo_preopt_dir):
            os.makedirs(self.not_geo_preopt_dir)
        self.log(f"Failed preoptimization geometry directory: {self.not_geo_preopt_dir}")

        self.geo_opt_dir = os.path.join(job_dir, "Geometries", "Optimized_Geo")
        if not os.path.exists(self.geo_opt_dir):
            os.makedirs(self.geo_opt_dir)
        self.log(f"DFT optimized geometry directory:          {self.geo_opt_dir}")

        self.not_geo_opt_dir = os.path.join(job_dir, "Geometries", "Not_Optimized_Geo")
        if not os.path.exists(self.not_geo_opt_dir):
            os.makedirs(self.not_geo_opt_dir)
        self.log(f"Failed DFT optimized geometry directory:   {self.not_geo_opt_dir}")

    def check_termination_succes(self, result: Results) -> bool:
        term = result.readrkf("General", "termination status", "ams")
        if term == "NORMAL TERMINATION":
            return True
        elif "NORMAL TERMINATION" in term:
            return "WARNING"
        return

    def oxidation_potential(
        self,
        molecule: Molecule,
        method: str = "screening",
        name: str = None,
        pre_optimize: bool = True,
        job_dir: str = os.getcwd(),
        COSMORS_solvent_path: str = None,
    ) -> float:

        assert method in [
            "DC",
            "TC-COSMO",
            "TC-COSMO-RS",
            "screening",
        ], 'Argument "method" must be "DC", "TC-COSMO", "TC-COSMO-RS" or "screening"'

        # set name
        if name is None:
            name = molecule.properties.name

        self.log("========================================================================")
        self.log(f"Starting oxidation potential calculation for molecule {name}:\n")

        # set paths for the jobs:
        self.log("Setting paths ...")
        self.set_paths(job_dir)

        self.log("\nInitial coordinates:")
        self.log(molecule)
        self.log("Settings:")
        self.log(f"\tName:              {name}")
        self.log(f"\tMethod:            {method}")
        self.log(f"\tPreoptimize:       {pre_optimize}")

        if pre_optimize:
            self.log("\nPre-optimizing the molecule ...", newline=False)
            molecule = self._pre_optimize(molecule, name=name)
            self.log(" done!")

        Gelectron = -0.0375
        # get on with the actual calculations
        if method == "DC":
            GO_os = self._calculation_step(
                molecule, task="GeometryOptimization", name=name, state="oxidized", phase="solvent"
            )
            GO_ns = self._calculation_step(
                molecule, task="GeometryOptimization", name=name, state="neutral", phase="solvent"
            )

            oxpot = GO_os["gibbs_energy"] - GO_ns["gibbs_energy"] + Gelectron

        elif method == "TC-COSMO":
            GO_nv = self._calculation_step(
                molecule, task="GeometryOptimization", name=name, state="neutral", phase="vacuum"
            )
            SP_nv_ns = self._calculation_step(
                GO_nv["geometry"], task="SinglePoint", name=name, state="neutral", phase="solvent", frequencies=False
            )
            GO_ns = self._calculation_step(
                molecule, task="GeometryOptimization", name=name, state="neutral", phase="solvent", frequencies=False
            )
            SP_nv_nv = self._calculation_step(
                GO_ns["geometry"], task="SinglePoint", name=name, state="neutral", phase="vacuum", frequencies=False
            )

            GO_ov = self._calculation_step(
                molecule, task="GeometryOptimization", name=name, state="oxidized", phase="vacuum"
            )
            SP_ov_os = self._calculation_step(
                GO_ov["geometry"], task="SinglePoint", name=name, state="oxidized", phase="solvent", frequencies=False
            )
            GO_os = self._calculation_step(
                molecule, task="GeometryOptimization", name=name, state="oxidized", phase="solvent", frequencies=False
            )
            SP_ov_ov = self._calculation_step(
                GO_os["geometry"], task="SinglePoint", name=name, state="oxidized", phase="vacuum", frequencies=False
            )

            oxidized_part = (
                GO_ov["gibbs_energy"] + SP_ov_os["dG_solvation"] + (SP_ov_ov["bond_energy"] - GO_ov["bond_energy"])
            )
            neutral_part = (
                GO_nv["gibbs_energy"] + SP_nv_ns["dG_solvation"] + (SP_nv_nv["bond_energy"] - GO_nv["bond_energy"])
            )
            oxpot = oxidized_part - neutral_part + Gelectron

        elif method == "TC-COSMO-RS":
            self.COSMORS_solvent_path = COSMORS_solvent_path
            assert os.path.exists(
                self.COSMORS_solvent_path
            ), f"Solvent database {self.COSMORS_solvent_path} does not exist"
            GO_nv = self._calculation_step(
                molecule, task="GeometryOptimization", name=name, state="neutral", phase="vacuum"
            )
            SP_nv_ns = self._calculation_step(
                GO_nv["geometry"], task="SinglePoint", name=name, state="neutral", use_COSMORS=True, frequencies=False
            )
            GO_ns = self._calculation_step(
                molecule, task="GeometryOptimization", name=name, state="neutral", use_COSMORS=True, frequencies=False
            )
            SP_ns_nv = self._calculation_step(
                GO_ns["geometry"], task="SinglePoint", name=name, state="neutral", phase="vacuum", frequencies=False
            )

            GO_ov = self._calculation_step(
                molecule, task="GeometryOptimization", name=name, state="oxidized", phase="vacuum"
            )
            SP_ov_os = self._calculation_step(
                GO_ov["geometry"], task="SinglePoint", name=name, state="oxidized", use_COSMORS=True, frequencies=False
            )
            GO_os = self._calculation_step(
                molecule, task="GeometryOptimization", name=name, state="oxidized", use_COSMORS=True, frequencies=False
            )
            SP_os_ov = self._calculation_step(
                GO_os["geometry"], task="SinglePoint", name=name, state="oxidized", phase="vacuum", frequencies=False
            )

            oxidized_part = (
                GO_ov["gibbs_energy"] + SP_ov_os["dG_solvation"] + (SP_os_ov["bond_energy"] - GO_ov["bond_energy"])
            )
            neutral_part = (
                GO_nv["gibbs_energy"] + SP_nv_ns["dG_solvation"] + (SP_ns_nv["bond_energy"] - GO_nv["bond_energy"])
            )
            oxpot = oxidized_part - neutral_part + Gelectron

        elif method == "screening":
            self.COSMORS_solvent_path = COSMORS_solvent_path
            assert os.path.exists(
                self.COSMORS_solvent_path
            ), f"Solvent database {self.COSMORS_solvent_path} does not exist"
            GO_nv = self._calculation_step(
                molecule,
                task="GeometryOptimization",
                name=name,
                state="neutral",
                phase="vacuum",
                use_dftb=True,
                frequencies=False,
            )
            COSMO_nv = self._calculation_step(
                GO_nv["geometry"], task="SinglePoint", name=name, state="neutral", use_COSMORS=True, frequencies=False
            )

            GO_ov = self._calculation_step(
                molecule,
                task="GeometryOptimization",
                name=name,
                state="oxidized",
                phase="vacuum",
                use_dftb=True,
                frequencies=False,
            )
            COSMO_ov = self._calculation_step(
                GO_ov["geometry"], task="SinglePoint", name=name, state="oxidized", use_COSMORS=True, frequencies=False
            )

            oxpot = COSMO_ov["gibbs_energy"] - COSMO_nv["gibbs_energy"] + Gelectron

        self.log(f"\nOxidation potential: {oxpot:.4f} eV")
        return oxpot

    def _pre_optimize(
        self,
        molecule: Molecule,
        name: str = None,
    ) -> Molecule:
        """Method used to pre-optimize the molecule using DFTB3"""
        job = AMSJob(molecule=molecule, settings=self.pre_optimize_defaults, name=name + "_preoptimization")
        res = job.run()

        # if the optimization is succesful we return the new molecule
        if self.check_termination_succes(res):
            molecule = res.get_main_molecule()
            molecule.write(os.path.join(self.geo_preopt_dir, name + ".xyz"))
        else:
            molecule.write(os.path.join(self.not_geo_preopt_dir, name + ".xyz"))

        return molecule

    def _get_settings(
        self,
        task: str = "GeometryOptimization",
        state: str = "neutral",
        phase: str = "vacuum",
        frequencies: bool = True,
        use_dftb: bool = False,
        use_COSMORS: bool = False,
    ) -> Settings:
        """
        Method that generates settings for jobs based on
        """

        assert state in ["neutral", "oxidized"], 'argument "state" must be "neutral" or "oxidized"'
        assert phase in ["vacuum", "solvent"], 'argument "phase" must be "vacuum" or "solvent"'
        assert task in [
            "GeometryOptimization",
            "SinglePoint",
        ], 'argument "task" must be "GeometryOptimization", "SinglePoint" or "COSMO"'

        if use_COSMORS:
            defaults = self.DFT_defaults.copy()

            solvation_block = {
                "surf": "Delley",
                "solv": "name=CRS cav0=0.0 cav1=0.0",
                "charged": "method=Conj corr",
                "c-mat": "Exact",
                "scf": "Var All",
                "radii": {
                    "H": 1.30,
                    "C": 2.00,
                    "N": 1.83,
                    "O": 1.72,
                    "F": 1.72,
                    "Si": 2.48,
                    "P": 2.13,
                    "S": 2.16,
                    "Cl": 2.05,
                    "Br": 2.16,
                    "I": 2.32,
                },
            }

            defaults.input.adf.solvation = solvation_block
        else:
            if use_dftb:
                defaults = self.DFTB_defaults.copy()
            else:
                defaults = self.DFT_defaults.copy()

            # load cosmo solvent
            if phase == "solvent" and not use_COSMORS:
                defaults.soft_update(self.COSMO_defaults)

        defaults.input.ams.task = task

        if frequencies:
            defaults.soft_update(self.frequencies_defaults)

        # handle state, if neutral the settings are already correct
        if state == "oxidized":
            if use_dftb:
                defaults.soft_update(self.DFTB_oxidized_defaults)
            else:
                defaults.soft_update(self.DFT_oxidized_defaults)

        return defaults

    def _COSMORS_property(self, solvent_path: str, solute_path: str, name: str, temperature: float = 298.15) -> float:
        """This method runs a COSMORS property job to obtain the activity coefficient
        which will also calculate G solute which we need to calculate the oxidation
        potential
        """

        defaults = Settings()
        defaults.input.property._h = "ACTIVITYCOEF"
        compounds = [Settings(), Settings()]
        compounds[0]._h = solvent_path
        compounds[1]._h = solute_path
        compounds[0].frac1 = 1
        compounds[1].frac1 = 0

        defaults.input.temperature = str(temperature)
        defaults.input.compound = compounds

        res = CRSJob(settings=defaults, name=name).run().get_results()
        if res:
            return float(Units.convert(res["G solute"][1], "kcal/mol", "hartree"))
        else:
            return False

    def _calculation_step(
        self,
        molecule: Molecule,
        task: str = "GeometryOptimization",
        state: str = "neutral",
        phase: str = "vacuum",
        frequencies: bool = True,
        use_dftb: bool = False,
        use_COSMORS: bool = False,
        name: str = None,
    ) -> dict:
        """Method used to optimize the geometry of molecule using DFT (by default B3LYP)
        Other settings may be supplied using settings which will be soft-updated using self.DFT_defaults
        State specifies whether the molecule is neutral or oxidised
        Phase specifies whether the system is in vacuum or solvated
        if use_dftb, the system will be optimised using DFTB (by default GFN1-xTB) instead of DFT
        """

        assert state in ["neutral", "oxidized"], 'argument "state" must be "neutral" or "oxidized"'
        assert phase in ["vacuum", "solvent"], 'argument "phase" must be "vacuum" or "solvent"'
        assert task in [
            "GeometryOptimization",
            "SinglePoint",
        ], 'argument "task" must be "GeometryOptimization" or "SinglePoint"'

        if use_COSMORS:
            phase = "solvent"

        settings = self._get_settings(
            task=task, use_dftb=use_dftb, use_COSMORS=use_COSMORS, state=state, phase=phase, frequencies=frequencies
        )

        # summarize job in one string
        task_abbrev = {"GeometryOptimization": "GO", "SinglePoint": "SP"}[task]
        job_desc = f"{task_abbrev}_{state}_{phase}"
        if use_COSMORS:
            job_desc += "_COSMO-RS"
        if use_dftb:
            job_desc += "_DFTB"

        self.log(f'\nStarting calculation {name + "_" + job_desc}')
        self.log(f"\ttask                 = {task}")
        self.log(f"\tuse_dftb             = {use_dftb}")
        self.log(f"\tuse_COSMORS          = {use_COSMORS}")
        self.log(f"\tfrequencies          = {frequencies}")
        self.log(f"\tstate                = {state}")
        self.log(f"\tphase                = {phase}")

        # run the job
        job = AMSJob(molecule=molecule, settings=settings, name=name + "_" + job_desc)
        res = job.run()

        result_dict = {}
        # pull out results
        if self.check_termination_succes(res):
            self.log(f"\tSuccessfull          = {self.check_termination_succes(res)}")  # True or WARNING
            # set some default values
            bond_energy = None
            gibbs_energy = None

            # If we are doing COSMO calculations then we need to run an additional job to obtain the activity coefficient
            # when calculating the activity coefficient, the G solute is also calculated.
            if use_COSMORS:
                resfile = KFFile(res["adf.rkf"])
                cosmo_data = resfile.read_section("COSMO")
                coskf = KFFile(os.path.join(job.path, job.name + ".coskf"))
                for k, v in cosmo_data.items():
                    coskf.write("COSMO", k, v)
                res.collect()
                bond_energy = res.readrkf("AMSResults", "Energy", "adf")
                gibbs_energy = self._COSMORS_property(
                    self.COSMORS_solvent_path, os.path.join(job.path, job.name + ".coskf"), job.name + "_ACTIVITYCOEF"
                )
            # if we dont use COSMO-RS we can just extract the Gibbs and bonding energies from the regular job
            else:
                if use_dftb:
                    bond_energy = res.readrkf("AMSResults", "Energy", "dftb")
                    if frequencies:
                        gibbs_energy = res.readrkf("Thermodynamics", "Gibbs free Energy", "dftb")
                else:
                    bond_energy = res.readrkf("Energy", "Bond Energy", "adf")
                    if frequencies:
                        gibbs_energy = res.readrkf("Thermodynamics", "Gibbs free Energy", "adf")

            self.log("\tResults:")
            if not bond_energy is None:
                result_dict["bond_energy"] = Units.convert(bond_energy, "hartree", "eV")
                self.log(f'\t\tBond Energy  = {result_dict["bond_energy"]:.4f} eV')
            if not gibbs_energy is None:
                result_dict["gibbs_energy"] = Units.convert(gibbs_energy, "hartree", "eV")
                self.log(f'\t\tGibbs Energy = {result_dict["gibbs_energy"]:.4f} eV')

            # extract also optimised molecule
            if task == "GeometryOptimization":
                opt_mol = res.get_main_molecule()
                opt_mol.write(os.path.join(self.geo_opt_dir, name + "_" + job_desc + ".xyz"))
                result_dict["geometry"] = res.get_main_molecule()

            # and if the phase is solvent we also need the solvation gibbs energy change
            if phase == "solvent":
                dG_solvation = res.readrkf("Energy", "Solvation Energy (el)", "adf") + res.readrkf(
                    "Energy", "Solvation Energy (cd)", "adf"
                )
                result_dict["dG_solvation"] = Units.convert(dG_solvation, "hartree", "eV")
                self.log(f'\t\tdG_solvation = {result_dict["dG_solvation"]:.4f} eV')

        else:
            self.log("\tSuccessfull          = False")
            if task == "GeometryOptimization":
                molecule.write(os.path.join(self.not_geo_opt_dir, name + "_" + job_desc + ".xyz"))

        return result_dict

    def log(self, line, newline=True):
        if not self.logfile is None:
            self.logfile.write(str(line) + "\n" * newline)
            self.logfile.flush()
        print(line)


if __name__ == "__main__":
    job_dir = "./Test"
    if not os.path.exists(job_dir):
        os.makedirs(job_dir)

    COSMORS_solvent_path = os.path.abspath("Dichloromethane.coskf")

    results = {}
    for mol_file in ["./molecules/NDI44.xyz"]:
        mol_name = os.path.basename(mol_file).split(".")[0]
        results[mol_name] = {}
        for method in ["screening", "TC-COSMO", "TC-COSMO-RS", "DC"]:
            job_name = None  # if set to None, a name will be generated

            if job_name is None:
                job_name = mol_name + "_" + method

            # calculation part
            init(path=job_dir, folder=job_name)

            workdir = config.default_jobmanager.workdir
            logfile = open(f"{workdir}/{job_name}_python.log", "w")

            OxPotCalc = OxidationPotentialCalculator(logfile=logfile)
            mol = Molecule(mol_file)
            oxpot = OxPotCalc(mol, job_dir=workdir, method=method, COSMORS_solvent_path=COSMORS_solvent_path)
            results[mol_name][method] = oxpot
            finish()

    print("Oxidation Potentials:")
    name_len = max(len("System"), max(len(n) for n in results))
    methods = list(results[list(results.keys())[0]])
    method_lens = [max(9, len(m)) for m in methods]
    print(f'{"System".ljust(name_len)} | {" | ".join([m.ljust(l) for m, l in zip(methods, method_lens)])}')
    for name, res in results.items():
        s = f'{name.ljust(name_len)} | {" | ".join([(str(round(res[m],3)) + " eV").rjust(l) for m, l in zip(methods, method_lens)])}'
        print(s)

    print("\nCalculations coomplete!\a")



#!/usr/bin/env amspython
import os

from ase.calculators.emt import EMT
from scm.plams import *


def get_calculator():
    """
    Return an ASE calculator for use with Engine ASE.

    The function must be called "get_calculator"

    """
    return EMT()


def main():
    init()

    mol = from_smiles("O")
    mol.lattice = [
        [
            3.0,
            0.0,
            0.0,
        ],
        [0.0, 3.0, 0.0],
        [0.0, 0.0, 3.0],
    ]

    s = Settings()
    s.runscript.nproc = 1
    s.input.ams.task = "GeometryOptimization"
    s.input.ams.GeometryOptimization.Convergence.Gradients = 0.01  # hartree/ang
    # get the get_calculator function, which returns an initialized ASE Calculator,
    # from the current file
    s.input.ASE.File = os.path.abspath(__file__)
    # to call get_calculator(my_argument = my_value) with arguments use:
    # s.input.ASE.Arguments._1 = "my_argument = my_value"

    job = AMSJob(settings=s, molecule=mol, name="ams_with_custom_ase_calculator")
    job.run()

    energy = job.results.get_energy(unit="eV")
    print(f"AMS with custom ASE calculator (Engine ASE), EMT potential: final energy {energy:.3f} eV")

    finish()


if __name__ == "__main__":
    main()



import os
import subprocess
import tempfile
from typing import Any, Dict, List, Literal, Optional, Tuple, Union, overload

import numpy as np
from scm.plams.core.errors import MoleculeError
from scm.plams.core.private import saferun
from scm.plams.interfaces.adfsuite.ams import AMSJob
from scm.plams.mol.molecule import Molecule

try:
    from scm.plams.interfaces.molecule.rdkit import readpdb, writepdb
except ImportError:
    pass

__all__ = [
    "packmol",
    "packmol_in_void",
    "packmol_on_slab",
    "packmol_microsolvation",
    "PackMolError",
]


def tolist(x):
    return x if isinstance(x, list) else [x]


class PackMolError(MoleculeError):
    pass


class PackMolStructure:
    def __init__(
        self,
        molecule: Molecule,
        n_molecules: Optional[int] = None,
        n_atoms: Optional[int] = None,
        box_bounds: Optional[List[float]] = None,
        density: Optional[float] = None,
        fixed: bool = False,
        sphere: bool = False,
    ):
        """

        Class representing a packmol structure.

        molecule: |Molecule|
            The molecule

        n_molecules: int
            The number of molecules to insert

        n_atoms: int
            An approximate number of atoms to insert

        box_bounds: list of float
            [xmin, ymin, zmin, xmax, ymax, zmax] in angstrom. The min values should all be 0, i.e. [0., 0., 0., xmax, ymax, zmax]

        density: float
            Density in g/cm^3

        fixed: bool
            Whether the structure should be fixed at its original coordinates.

        sphere: bool
            Whether the molecules should be packed in a sphere. The radius is determined by getting the volume from the box bounds! Cannot be combined with ``fixed`` (``fixed`` takes precedence).

        """
        self.molecule = molecule
        if fixed:
            assert n_molecules is None or n_molecules == 1
            assert density is None
            self.n_molecules = 1
            if molecule.lattice and len(molecule.lattice) == 3:
                self.box_bounds = [
                    0.0,
                    0.0,
                    0.0,
                    molecule.lattice[0][0],
                    molecule.lattice[1][1],
                    molecule.lattice[2][2],
                ]
            else:
                self.box_bounds = None
            self.fixed = True
            self.sphere = False
        else:
            if box_bounds and density:
                if n_molecules or n_atoms:
                    raise ValueError("Cannot set all n_molecules or n_atoms together with (box_bounds AND density)")
                n_molecules = self._get_n_molecules_from_density_and_box_bounds(self.molecule, box_bounds, density)
            assert n_molecules is not None or n_atoms is not None
            if n_molecules is None:
                self.n_molecules = self._get_n_molecules(self.molecule, n_atoms)
            else:
                self.n_molecules = n_molecules
            assert box_bounds or density
            self.box_bounds = box_bounds or self._get_box_bounds(self.molecule, self.n_molecules, density)
            self.fixed = False
            self.sphere = sphere

    def _get_n_molecules_from_density_and_box_bounds(self, molecule: Molecule, box_bounds: List[float], density: float):
        """density in g/cm^3"""
        molecule_mass = molecule.get_mass(unit="g")
        volume_ang3 = self.get_volume(box_bounds)
        volume_cm3 = volume_ang3 * 1e-24
        n_molecules = int(density * volume_cm3 / molecule_mass)
        return n_molecules

    def get_volume(self, box_bounds=None):
        bb = box_bounds or self.box_bounds
        vol = (bb[3] - bb[0]) * (bb[4] - bb[1]) * (bb[5] - bb[2])
        return vol

    def _get_n_molecules(self, molecule: Molecule, n_atoms: int):
        return n_atoms // len(molecule)

    def _get_box_bounds(self, molecule: Molecule, n_molecules: int, density: float):
        mass = n_molecules * molecule.get_mass(unit="g")
        volume_cm3 = mass / density
        volume_ang3 = volume_cm3 * 1e24
        side_length = volume_ang3 ** (1 / 3.0)
        return [0.0, 0.0, 0.0, side_length, side_length, side_length]

    def get_input_block(self, fname, tolerance):
        if self.n_molecules == 0 and not self.fixed:
            return ""
        if self.fixed:
            ret = f"""
            structure {fname}
            number 1
            fixed 0. 0. 0. 0. 0. 0.
            avoid_overlap yes
            end structure
            """
        elif self.sphere:
            vol = self.get_volume()
            # vol = 4*pi*r^3 /3
            # radius = (3*vol/(4*pi))**0.33333
            radius = (3 * vol / (4 * 3.14159)) ** 0.3333
            ret = f"""
            structure {fname}
              number {self.n_molecules}
              inside sphere 0. 0. 0. {radius}
            end structure
            """
        else:
            box_string = f"{self.box_bounds[0]+tolerance/2} {self.box_bounds[1]+tolerance/2} {self.box_bounds[2]+tolerance/2} {self.box_bounds[3]-tolerance/2} {self.box_bounds[4]-tolerance/2} {self.box_bounds[5]-tolerance/2}"
            ret = f"""
            structure {fname}
              number {self.n_molecules}
              inside box {box_string}
            end structure

        """
        return ret


class PackMol:
    def __init__(
        self,
        tolerance=2.0,
        structures: Optional[List[PackMolStructure]] = None,
        filetype="xyz",
        executable=None,
    ):
        """
        Class for setting up and running packmol.

        tolerance: float
            The packmol tolerance (approximate minimum interatomic distance)

        structures: list of PackMolStructure
            Structures to insert

        filetype: str
            One of 'xyz' or 'pdb'. Specifies the file format to use with packmol. 'pdb' requires rdkit.

        executable: str
            Path to the packmol executable. If not specified, $AMSBIN/packmol.exe will be used.

        Note: users are not recommended to use this class directly, but
        instead use the ``packmol``, ``packmol_on_slab`` and ``packmol_microsolvation``
        functions.

        """
        self.tolerance = tolerance
        self.structures = structures or []
        self.filetype = filetype
        self.executable = executable or os.path.join(os.path.expandvars("$AMSBIN"), "packmol.exe")
        if not os.path.exists(self.executable):
            raise RuntimeError("PackMol exectuable not found: " + self.executable)

    def add_structure(self, structure: PackMolStructure):
        self.structures.append(structure)

    def _get_complete_box_bounds(self) -> Tuple[float, float, float, float, float, float]:
        min_x = min(s.box_bounds[0] for s in self.structures if s.box_bounds is not None)
        min_y = min(s.box_bounds[1] for s in self.structures if s.box_bounds is not None)
        min_z = min(s.box_bounds[2] for s in self.structures if s.box_bounds is not None)
        max_x = min(s.box_bounds[3] for s in self.structures if s.box_bounds is not None)
        max_y = min(s.box_bounds[4] for s in self.structures if s.box_bounds is not None)
        max_z = min(s.box_bounds[5] for s in self.structures if s.box_bounds is not None)

        # return min_x, min_y, min_z, max_x+self.tolerance, max_y+self.tolerance, max_z+self.tolerance
        return min_x, min_y, min_z, max_x, max_y, max_z

    def _get_complete_lattice(self) -> List[List[float]]:
        """
        returns a 3x3 list using the smallest and largest x/y/z box_bounds for all structures
        """
        if any(s.sphere for s in self.structures):
            return []
        (
            min_x,
            min_y,
            min_z,
            max_x,
            max_y,
            max_z,
        ) = self._get_complete_box_bounds()
        return [
            [max_x - min_x, 0.0, 0.0],
            [0.0, max_y - min_y, 0.0],
            [0.0, 0.0, max_z - min_z],
        ]

    def _get_complete_radius(self) -> float:
        """
        Calculates radius of sphere with the same volume as the
        cuboid from the box bounds

        :return: Radius in angstrom
        :rtype: float
        """
        volume = self._get_complete_volume()
        radius = (3 * volume / (4 * 3.14159)) ** 0.3333

        return radius

    def _get_complete_volume(self) -> float:
        """Returns volume based on box bounds in ang^3

        :return: Volume in ang^3
        :rtype: float
        """
        (
            min_x,
            min_y,
            min_z,
            max_x,
            max_y,
            max_z,
        ) = self._get_complete_box_bounds()

        volume = (max_x - min_x) * (max_y - min_y) * (max_z - min_z)

        return volume

    def run(self):
        """
        returns: a Molecule with the packed structures
        """

        output_molecule = Molecule()
        with tempfile.TemporaryDirectory() as tmpdir:
            output_fname = os.path.join(tmpdir, "output.xyz")
            input_fname = os.path.join(tmpdir, "input.inp")
            with open(input_fname, "w") as input_file:
                input_file.write(f"tolerance {self.tolerance}\n")
                input_file.write(f"filetype {self.filetype}\n")
                input_file.write(f"output {output_fname}\n")

                for i, structure in enumerate(self.structures):
                    structure_fname = os.path.join(tmpdir, f"structure{i}.{self.filetype}")
                    if self.filetype == "pdb":
                        with open(structure_fname, "w") as f:
                            writepdb(structure.molecule, f)
                    else:
                        structure.molecule.write(structure_fname)
                    input_file.write(structure.get_input_block(structure_fname, tolerance=self.tolerance))

            my_input = open(input_fname, "r")
            saferun(self.executable, stdin=my_input, stdout=subprocess.DEVNULL)
            my_input.close()

            if not os.path.exists(output_fname):
                raise PackMolError("Packmol failed. It may work if you try a lower density.")

            if self.filetype == "pdb":
                with open(output_fname, "r") as f:
                    output_molecule = readpdb(f)
            else:
                output_molecule = Molecule(output_fname)  # without periodic boundary conditions

            output_molecule.lattice = self._get_complete_lattice()

        return output_molecule


@overload
def packmol(
    molecules: Union[List[Molecule], Molecule],
    mole_fractions: Optional[List[float]] = ...,
    density: Optional[float] = ...,
    n_atoms: Optional[int] = ...,
    box_bounds: Optional[List[float]] = ...,
    n_molecules: Union[List[int], int, None] = ...,
    sphere: bool = ...,
    fix_first: bool = ...,
    keep_bonds: bool = ...,
    keep_atom_properties: bool = ...,
    region_names: Union[List[str], str, None] = ...,
    return_details: Literal[False] = ...,
    tolerance: float = ...,
    executable: Optional[str] = ...,
) -> Molecule: ...


@overload
def packmol(
    molecules: Union[List[Molecule], Molecule],
    mole_fractions: Optional[List[float]] = ...,
    density: Optional[float] = ...,
    n_atoms: Optional[int] = ...,
    box_bounds: Optional[List[float]] = ...,
    n_molecules: Union[List[int], int, None] = ...,
    sphere: bool = ...,
    fix_first: bool = ...,
    keep_bonds: bool = ...,
    keep_atom_properties: bool = ...,
    region_names: Union[List[str], str, None] = ...,
    return_details: Literal[True] = ...,
    tolerance: float = ...,
    executable: Optional[str] = ...,
) -> Tuple[Molecule, Dict[str, Any]]: ...


def packmol(
    molecules: Union[List[Molecule], Molecule],
    mole_fractions: Optional[List[float]] = None,
    density: Optional[float] = None,
    n_atoms: Optional[int] = None,
    box_bounds: Optional[List[float]] = None,
    n_molecules: Union[List[int], int, None] = None,
    sphere: bool = False,
    fix_first: bool = False,
    keep_bonds: bool = True,
    keep_atom_properties: bool = True,
    region_names: Union[List[str], str, None] = None,
    return_details: bool = False,
    tolerance: float = 2.0,
    executable: Optional[str] = None,
) -> Union[Molecule, Tuple[Molecule, Dict[str, Any]]]:
    """
    Create a fluid of the given ``molecules``. The function will use the
    given input parameters and try to obtain good values for the others. You *must*
    specify ``density`` and/or ``box_bounds``.

    molecules : |Molecule| or list of Molecule
        The molecules to pack

    mole_fractions : list of float
        The mole fractions (in the same order as ``molecules``). Cannot be combined with ``n_molecules``. If not given, an equal (molar) mixture of all components will be created.

    density: float
        The total density (in g/cm^3) of the fluid

    n_atoms: int
        The (approximate) number of atoms in the final mixture

    box_bounds: list of float (length 6)
        The box in which to pack the molecules. The box is orthorhombic and should be specified as [xmin, ymin, zmin, xmax, ymax, zmax]. The minimum values should all be set to 0, i.e. set box_bounds=[0., 0., 0., xmax, ymax, zmax]. If not specified, a cubic box of appropriate dimensions will be used.

    n_molecules : int or list of int
        The (exact) number of molecules for each component (in the same order as ``molecules``). Cannot be combined with ``mole_fractions``.

    sphere: bool
        Whether the molecules should be packed in a sphere. The radius is determined by getting the volume from the box bounds!

    fix_first: bool
        Whether to keep the first molecule fixed. This can only be used with ``n_molecules=[1, ..., ...]``. Defaults to False.

    keep_bonds : bool
        If True, the bonds from the constituent molecules will be kept in the returned Molecule

    keep_atom_properties : bool
        If True, the atom.properties (e.g. force-field atom types) of the constituent molecules will be kept in
        the returned Molecule

    region_names : str or list of str
        Populate the region information for each atom. Should have the same length and order as ``molecules``.
        By default the regions are named ``mol0``, ``mol1``, etc.

    tolerance: float
        The packmol tolerance (approximately the minimum intermolecular distance). When packing
        a periodic box, half the tolerance will be excluded from each face of the box.

    return_details : bool
        Return a 2-tuple (Molecule, dict) where the dict has keys like 'n_molecules', 'mole_fractions', 'density',
        etc. They contain the actual details of the returned molecule, which may differ slightly from
        the requested quantities.

        Returned keys:

        * 'n_molecules': list of integer with actually added number of molecules
        * 'mole_fractions': list of float with actually added mole fractions
        * 'density': float, gives the density in g/cm^3
        * 'n_atoms': int, the number of atoms in the returned molecule
        * 'molecule_type_indices': list of int of length n_atoms. For each atom, give an integer index for which TYPE of molecule it belongs to.
        * 'molecule_indices': list of int of length n_atoms. For each atom, give an integer index for which molecule it belongs to
        * 'atom_indices_in_molecule': list of int of length n_atoms. For each atom, give an integer index for which position in the molecule it is.
        * 'volume': float. The volume of the bounding box / packed sphere in ang^3.

    executable : str
        The path to the packmol executable. If not specified, ``$AMSBIN/packmol.exe`` will be used (which is the correct path for the Amsterdam Modeling Suite).

    Useful combinations:

    * ``mole_fractions``, ``density``, ``n_atoms``: Create a mixture with a given density and approximate number of atoms (a cubic box will be created)

    * ``mole_fractions``, ``density``, ``box_bounds``: Create a mixture with a given density inside a given box (the number of molecules will approximately match the density and mole fractions)

    * ``n_molecules``, ``density``: Create a mixture with the given number of molecules and density (a cubic box will be created)

    * ``n_molecules``, ``box_bounds``: Create a mixture with the given number of molecules inside the given box

    Example:

    .. code-block:: python

        packmol(molecules=[from_smiles('O'), from_smiles('C')],
                mole_fractions=[0.8, 0.2],
                density=0.8,
                n_atoms=100)

    Returns: a |Molecule| or tuple (Molecule, dict)
        If return_details=False, return a Molecule. If return_details=True, return a tuple.


    """
    # Input arguments allow for lots of combinations.
    # Let's try to check that the specified combination makes sense ...
    if n_atoms is None and n_molecules is None and density is None:
        raise ValueError("Illegal combination of arguments: must specify either n_atoms, n_molecules or density")
    if n_atoms is not None and n_molecules is not None:
        raise ValueError("Illegal combination of arguments: n_atoms and n_molecules are mutually exclusive")
    if density is None and box_bounds is None:
        raise ValueError("Illegal combination of arguments: must specify either density or box_bounds")
    if n_atoms is not None and box_bounds is not None and density is not None:
        raise ValueError("Illegal combination of arguments: n_atoms, box_bounds and density specified at the same time")
    if n_molecules is not None and box_bounds is not None and density is not None:
        raise ValueError(
            "Illegal combination of arguments: n_molecules, box_bounds and density specified at the same time"
        )
    if mole_fractions is not None and n_molecules is not None:
        raise ValueError("Illegal combination of arguments: mole_fractions and n_molecules are mutually exclusive")
    if fix_first:
        if n_molecules is None or np.isscalar(n_molecules) or n_molecules[0] != 1:
            raise ValueError(
                f"Illegal combination of arguments: fix_first requires that n_molecules is a list where the first element is 1. Received n_molecules={n_molecules}"
            )
    if isinstance(molecules, list):
        if n_molecules is not None:
            if not isinstance(n_molecules, list):
                raise ValueError("Illegal combination of arguments: molecules is a list, but n_molecules is not")
            if len(n_molecules) != len(molecules):
                raise ValueError("Illegal combination of arguments: len(n_molecules) != len(molecules)")
        if mole_fractions is not None:
            if not isinstance(mole_fractions, list):
                raise ValueError("Illegal combination of arguments: molecules is a list, but mole_fractions is not")
            if len(mole_fractions) != len(molecules):
                raise ValueError("Illegal combination of arguments: len(mole_fractions) != len(molecules)")
        if region_names is not None:
            if not isinstance(region_names, list):
                raise ValueError("Illegal combination of arguments: molecules is a list, but region_names is not")
            if len(region_names) != len(molecules):
                raise ValueError("Illegal combination of arguments: len(region_names) != len(molecules)")
    else:
        if n_molecules is not None and isinstance(n_molecules, list):
            raise ValueError("Illegal combination of arguments: n_molecules is a list, when molecules is not")
        if mole_fractions is not None and isinstance(mole_fractions, list):
            raise ValueError("Illegal combination of arguments: mole_fractions is a list, when molecules is not")
        if region_names is not None and isinstance(region_names, list):
            raise ValueError("Illegal combination of arguments: region_names is a list, when molecules is not")

    molecules = tolist(molecules)
    if mole_fractions is None:
        mole_fractions = [1.0 / len(molecules)] * len(molecules)

    if n_molecules:
        n_molecules = tolist(n_molecules)
        sum_n_molecules = np.sum(n_molecules)
        if np.isclose(sum_n_molecules, 0):
            raise ValueError(
                f"The sum of n_molecules is {sum_n_molecules}, which is very close to 0. "
                f"Specify larger numbers. n_molecules specified: {n_molecules}"
            )
        if any(x < 0 for x in n_molecules):
            raise ValueError(f"All n_molecules must be >= 0. " f"n_molecules specified: {n_molecules}")

    xs = np.array(mole_fractions)
    sum_xs = np.sum(xs)
    if np.isclose(sum_xs, 0):
        raise ValueError(
            f"The sum of mole fractions is {sum_xs}, which is very close to 0. "
            f"Specify larger numbers. Mole fractions specified: {xs}"
        )
    if np.any(xs < 0):
        raise ValueError(f"All mole fractions must be >= 0. " f"Mole fractions specified: {mole_fractions}")

    atoms_per_mol = np.array([len(a) for a in molecules])
    masses = np.array([m.get_mass(unit="g") for m in molecules])

    coeffs = None

    if n_molecules:
        coeffs = np.int_(n_molecules)
    elif n_atoms:
        coeff_0 = n_atoms / np.dot(xs, atoms_per_mol)
        coeffs_floats = xs * coeff_0
        coeffs = np.int_(np.round(coeffs_floats))

    if (n_atoms or n_molecules) and density and not box_bounds:
        mass = np.dot(coeffs, masses)
        volume_cm3 = mass / density
        volume_ang3 = volume_cm3 * 1e24
        side_length = volume_ang3 ** (1 / 3.0)
        box_bounds = [0.0, 0.0, 0.0, side_length, side_length, side_length]
    elif box_bounds and density and not n_molecules:
        volume_cm3 = (
            (box_bounds[3] - box_bounds[0]) * (box_bounds[4] - box_bounds[1]) * (box_bounds[5] - box_bounds[2]) * 1e-24
        )
        mass_g = volume_cm3 * density
        coeffs = mass_g / np.dot(xs, masses)
        coeffs = xs * coeffs
        coeffs = np.int_(np.round(coeffs))

    if coeffs is None:
        raise ValueError(
            f"Illegal combination of arguments: n_atoms={n_atoms}, n_molecules={n_molecules}, box_bounds={box_bounds}, density={density}"
        )

    pm = PackMol(executable=executable, tolerance=tolerance)
    if sphere and len(molecules) == 2 and n_molecules and n_molecules[0] == 1:
        # Special case used by packmol_microsolvation
        pm.add_structure(
            PackMolStructure(molecules[0], n_molecules[0], box_bounds=box_bounds, sphere=False, fixed=True)
        )
        pm.add_structure(
            PackMolStructure(molecules[1], n_molecules[1], box_bounds=box_bounds, sphere=True, fixed=False)
        )
    else:
        for i, (mol, n_mol) in enumerate(zip(molecules, coeffs)):
            if fix_first and i == 0:
                pm.add_structure(
                    PackMolStructure(mol, n_molecules=n_mol, box_bounds=box_bounds, sphere=False, fixed=True)
                )
            else:
                pm.add_structure(PackMolStructure(mol, n_molecules=n_mol, box_bounds=box_bounds, sphere=sphere))

    out = pm.run()

    # packmol returns the molecules sorted
    molecule_type_indices = []  # [0,0,0,...,1,1,1] # two different molecules with 3 and 5 atoms
    molecule_indices = (
        []
    )  # [0,0,0,1,1,1,2,2,2,....,58,58,58,58,58,59,59,59,59,59] # two different molecules with 3 and 5 atoms
    atom_indices_in_molecule = []  # [0,1,2,0,1,2,...,0,1,2,3,4,0,1,2,3,4]
    current = 0
    for i, (mol, n_mol) in enumerate(zip(molecules, coeffs)):
        molecule_type_indices += [i] * n_mol * len(mol)
        atom_indices_in_molecule += list(range(len(mol))) * n_mol

        temp = list(range(current, current + n_mol))
        molecule_indices += list(np.repeat(temp, len(mol)))
        current += n_mol
    assert len(molecule_type_indices) == len(out)
    assert len(molecule_indices) == len(out)
    assert len(atom_indices_in_molecule) == len(out)

    try:
        volume = out.unit_cell_volume(unit="angstrom")
        density = out.get_density() * 1e-3  # g / cm^3
    except ValueError:  # not periodic, presumably when sphere=True
        volume = pm._get_complete_volume()
        mass = out.get_mass(unit="g")
        density = mass / (volume * 1e-24)  # g / cm^3
    details = {
        "n_molecules": coeffs.tolist(),
        "mole_fractions": (coeffs / np.sum(coeffs)).tolist() if np.sum(coeffs) > 0 else [0.0] * len(coeffs),
        "n_atoms": len(out),
        "molecule_type_indices": molecule_type_indices,  # for each atom, indicate which type of molecule it belongs to by an integer index (starts with 0)
        "molecule_indices": molecule_indices,  # for each atoms, indicate which molecule it belongs to by an integer index (starts with 0)
        "atom_indices_in_molecule": atom_indices_in_molecule,
        "volume": volume,
        "density": density,
    }

    if sphere:
        details["radius"] = (volume * 3 / (4 * 3.1415926535)) ** (1 / 3.0)

    if keep_atom_properties:
        for at, molecule_type_index, atom_index_in_molecule in zip(
            out, molecule_type_indices, atom_indices_in_molecule
        ):
            at.properties = molecules[molecule_type_index][atom_index_in_molecule + 1].properties.copy()

    if keep_bonds:
        out.delete_all_bonds()
        for imol, mol in enumerate(molecules):
            for b in mol.bonds:
                i1, i2 = sorted(mol.index(b))  # 1-based
                for iout, (molecule_type, atom_index_molecule) in enumerate(
                    zip(molecule_type_indices, atom_indices_in_molecule)
                ):
                    if molecule_type != imol:
                        continue
                    if i1 != atom_index_molecule + 1:
                        continue
                    new_i1 = iout + 1  # iout 0-based
                    new_i2 = iout + 1 + i2 - i1  # iout 0-based
                    out.add_bond(out[new_i1], out[new_i2], order=b.order)

    if region_names:
        region_names = tolist(region_names)
    else:
        region_names = [f"mol{i}" for i in range(len(molecules))]

    for at, molindex in zip(out, molecule_type_indices):
        AMSJob._add_region(at, region_names[molindex])

    tot_charge = sum(int(mol.properties.get("charge", 0)) * c for mol, c in zip(molecules, coeffs))
    if tot_charge != 0:
        out.properties.charge = tot_charge

    if return_details:
        return out, details

    return out


def get_packmol_solid_liquid_box_bounds(slab: Molecule):
    slab_max_z = max(at.coords[2] for at in slab)
    slab_min_z = min(at.coords[2] for at in slab)
    liquid_min_z = slab_max_z
    liquid_max_z = liquid_min_z + slab.lattice[2][2] - (slab_max_z - slab_min_z)
    box_bounds = [
        0.0,
        0.0,
        liquid_min_z + 1.5,
        slab.lattice[0][0],
        slab.lattice[1][1],
        liquid_max_z - 1.5,
    ]
    return box_bounds


def packmol_in_void(
    host: Molecule,
    molecules: Union[List[Molecule], Molecule],
    n_molecules: Union[List[int], int],
    keep_bonds: bool = True,
    keep_atom_properties: bool = True,
    region_names: Optional[List[str]] = None,
    tolerance: float = 2.0,
    return_details: bool = False,
    executable: Optional[str] = None,
):
    """
    Pack molecules inside voids in a crystal.

    host: Molecule
        The host molecule. Must be 3D-periodic and the cell must be orthorhombic (all angles 90 degrees) with the lattice vectors parallel to the cartesian axes (all off-diagonal components must be 0).

    For the other arguments, see the ``packmol`` function.

    Note: ``region_names`` needs to have one more element than the list of
    ``molecules``. For example ``region_names=['host', 'guest1',
    'guest2']``.

    """
    if len(host.lattice) != 3:
        raise ValueError("host in packmol_in_void must be 3D periodic")
    if host.cell_angles() != [90.0, 90.0, 90.0]:
        raise ValueError("host in packmol_in_void must be have orthorhombic cell")

    my_host = host.copy()
    my_host.map_to_central_cell(around_origin=False)
    box_bounds = [0.0, 0.0, 0.0, my_host.lattice[0][0], my_host.lattice[1][1], my_host.lattice[2][2]]

    my_molecules = [my_host] + tolist(molecules)
    my_n_molecules = [1] + tolist(n_molecules)

    ret = packmol(
        molecules=my_molecules,
        n_molecules=my_n_molecules,
        box_bounds=box_bounds,
        keep_bonds=keep_bonds,
        keep_atom_properties=keep_atom_properties,
        region_names=region_names,
        return_details=return_details,
        fix_first=True,
        tolerance=tolerance,
        executable=executable,
    )

    return ret


def packmol_on_slab(
    slab: Molecule,
    molecules: Union[List[Molecule], Molecule],
    mole_fractions: Optional[List[float]] = None,
    density: float = 1.0,
    keep_bonds: bool = True,
    keep_atom_properties: bool = True,
    region_names: Optional[List[str]] = None,
    executable: Optional[str] = None,
):
    """

    Creates a solid/liquid interface with an approximately correct density. The
    density is calculated for the volume not occupied by the slab (+ 1.5
    angstrom buffer at each side of the slab).

    Returns: a |Molecule|

    slab : |Molecule|
        The system must have a 3D lattice (including a vacuum gap along z) and be orthorhombic. The vacuum gap will be filled with the liquid.

    For the other arguments, see ``packmol``.

    Example:

    .. code-block:: python

        packmol_on_slab(slab=slab_3d_with_vacuum_gap,
                        molecules=[from_smiles('O'), from_smiles('C')],
                        mole_fractions=[0.8, 0.2],
                        density=0.8)

    """
    if len(slab.lattice) != 3:
        raise ValueError("slab in packmol_on_slab must be 3D periodic: slab in xy-plane with vacuum gap along z-axis")
    if not all(np.isclose(slab.cell_angles(), [90.0, 90.0, 90.0])):
        raise ValueError("slab in packmol_on_slab must be have orthorhombic cell")

    liquid = packmol(
        molecules=molecules,
        mole_fractions=mole_fractions,
        density=density,
        box_bounds=get_packmol_solid_liquid_box_bounds(slab),
        keep_bonds=keep_bonds,
        keep_atom_properties=keep_atom_properties,
        region_names=region_names,
        executable=executable,
    )

    # Map all liquid molecules to [0..1]
    # NOTE: We need to be using the lattice of the slab for this!
    #       The lattice of the liquid is different ...
    liquid.lattice = slab.lattice
    # If the slab has cell-shifts for the bonds, the liquid also needs to have
    # them. If if would not have cell-shifts, they would not be updated in the
    # map_to_central_cell call, even though they would become significant when
    # combining with the slab that has them: minimum image convention is only
    # assumed if no bond has cell-shifts.
    if liquid.bonds and any(b.has_cell_shifts() for b in slab.bonds):
        for b in liquid.bonds:
            b.properties.suffix = "0 0 0"
    liquid.map_to_central_cell(around_origin=False)
    if liquid.bonds and any(b.has_cell_shifts() for b in slab.bonds):
        for b in liquid.bonds:
            if b.properties.suffix == "0 0 0":
                del b.properties.suffix

    # Shift liquid molecules (now in [0..1]) on top of the slab.
    # The slab could be anywhere, e.g. [-0.5..0.5] ...
    slab_center_x = (max(at.coords[0] for at in slab) + min(at.coords[0] for at in slab)) / 2
    slab_center_y = (max(at.coords[1] for at in slab) + min(at.coords[1] for at in slab)) / 2
    liquid_center_x = (max(at.coords[0] for at in liquid) + min(at.coords[0] for at in liquid)) / 2
    liquid_center_y = (max(at.coords[1] for at in liquid) + min(at.coords[1] for at in liquid)) / 2
    liquid.translate([-liquid_center_x + slab_center_x, -liquid_center_y + slab_center_y, 0.0])

    out = slab.copy()
    for at in out:
        AMSJob._add_region(at, "slab")
    out.add_molecule(liquid)
    return out


def get_n_from_density_and_box_bounds(molecule, box_bounds, density):
    molecule_mass = molecule.get_mass(unit="g")
    volume_ang3 = (box_bounds[3] - box_bounds[0]) * (box_bounds[4] - box_bounds[1]) * (box_bounds[5] - box_bounds[2])
    volume_cm3 = volume_ang3 * 1e-24
    n_molecules = int(density * volume_cm3 / molecule_mass)
    return n_molecules


def packmol_microsolvation(
    solute: Molecule,
    solvent: Molecule,
    density: float = 1.0,
    threshold: float = 3.0,
    keep_bonds: bool = True,
    keep_atom_properties: bool = True,
    region_names: List[str] = ["solute", "solvent"],
    executable: Optional[str] = None,
):
    """
    Microsolvation of a ``solute`` with a ``solvent`` with an approximate ``density``.

    solute: |Molecule|
        The solute to be surrounded by solvent molecules

    solvent: |Molecule|
        The solvent molecule

    density: float
        Approximate density in g/cm^3

    threshold: float
        Distance in angstrom. Any solvent molecule for which at least 1 atom is within this threshold to the solute molecule will be kept

    For the other arguments, see ``packmol``.

    """

    solute_coords = solute.as_array()
    com = np.mean(solute_coords, axis=0)
    plams_solute = solute.copy()
    plams_solute.translate(-com)
    solute_coords = plams_solute.as_array()
    box_bounds = [0.0, 0.0, 0.0] + list(np.max(solute_coords, axis=0) - np.min(solute_coords, axis=0) + 3 * threshold)

    n_solvent = get_n_from_density_and_box_bounds(solvent, box_bounds, density=density)

    plams_solvated = packmol(
        [plams_solute, solvent],
        n_molecules=[1, n_solvent],
        box_bounds=box_bounds,
        keep_bonds=keep_bonds,
        keep_atom_properties=keep_atom_properties,
        region_names=region_names,
        sphere=True,
        executable=executable,
    )

    solute_indices = [i for i, at in enumerate(plams_solvated, 1) if i <= len(solute)]
    newmolecule = plams_solvated.get_complete_molecules_within_threshold(solute_indices, threshold=threshold)

    return newmolecule



import dill as pickle
import pytest

from scm.plams.interfaces.adfsuite.ams import AMSJob
from scm.plams.core.settings import Settings


def test_pickle():
    job = AMSJob()
    pickle_bytes = pickle.dumps(job)
    job2 = pickle.loads(pickle_bytes)
    assert isinstance(job2, AMSJob)


def test_pickle_settings():
    s = Settings()
    s.input.ams.Task = "GeometryOptimization"
    s.input.ams.Properties.NormalModes = "Yes"
    s.input.DFTB.Model = "GFN1-xTB"
    job = AMSJob(settings=s)
    pickle_bytes = pickle.dumps(job)
    job2 = pickle.loads(pickle_bytes)
    assert isinstance(job2, AMSJob)
    assert job2.settings == job.settings


def test_pickle_pisa():
    try:
        from scm.input_classes.drivers import AMS
        from scm.input_classes.engines import DFTB

    except ImportError:
        pytest.skip("Skipping test because optional 'scm.pisa' package is not available")
    driver = AMS()
    driver.Task = "GeometryOptimization"
    driver.Properties.NormalModes = True
    driver.Engine = DFTB()
    driver.Engine.Model = "GFN1-xTB"
    job = AMSJob(settings=driver)
    pickle_bytes = pickle.dumps(job)
    job2 = pickle.loads(pickle_bytes)
    assert isinstance(job2, AMSJob)
    assert job2.settings == job.settings



from scm.plams.interfaces.adfsuite.ams import AMSJob
from scm.plams.tools.kftools import KFFile
from scm.plams.mol.molecule import Molecule
from scm.plams.core.basejob import MultiJob
from scm.plams.core.results import Results
from scm.plams.core.settings import Settings
from scm.conformers import ConformersJob
from scm.plams.recipes.adfcosmorscompound import ADFCOSMORSCompoundJob
import os


class ADFCOSMORSConfFilter:
    """
    This class allows the user to specify criteria on which to filter sets of conformers.

    Args:
        max_num_confs : The maximum number of conformers to bring forward to the next step.  These are automatically ordered by energy, so the first *n* lowest energy structures are carried carried forward to the next step.
        max_energy_range : The maximum allowable difference (in kcal/mol) between the lowest energy conformer and the highest energy conformer that will be carried forward.  For example, for a max_energy_range of 2 kcal/mol and a lowest energy structure of 1 kcal/mol, any structure with an energy above 3 (1+2) kcal/mol will be filtered out.
    """

    def __init__(self, max_num_confs=None, max_energy_range=None):
        self.max_num_confs = max_num_confs
        self.max_energy_range = max_energy_range


class ADFCOSMORSConfResults(Results):
    pass


class ADFCOSMORSConfJob(MultiJob):
    """
    This class allows for the user to implement a custom workflow for generating multiple conformers for use with COSMO-RS.  The class allows the user to input conformer generation strategies, job types to refine the energies/geometries, and filters to remove conformers between each calculation step.

    Args:
        molecule (|Molecule|) : A plams Molecule instance

    Keyword Args:
        conf_gen (ConformersJob) : A ConformersJob instance
        first_filter (ADFCOSMORSConfFilter) : The ADFCOSMORSConfFilter to apply to the initial set of sampled conformers
        additional : A list of (|Settings|, ADFCOSMORSConfFilter) tuples.  The elements of this list represent additional calculation steps (e.g., progressively higher levels of theory) and possibly different filters to apply at each step.
        final_filter (ADFCOSMORSConfFilter) : The ADFCOSMORSConfFilter to apply to the results of the gas phase ADF calculation for the conformers
        adf_singlepoint (bool) : A boolean indicating if the adf gas phase calculation in the cosmo-rs compound task should be a single point.  This defaults to False.
        initial_conformers : the (integer) number of initially sampled conformers.  This is only applied if the default conformer generation strategy is used.
        coskf_dir : a directory to put all the .coskf files generated for the conformers.  If this keyword is not specified, the .coskf files are put in the directory containing the adf gas phase calculation results
        coskf_name : a base name to be used with conformers.  All conformers will have the name *coskf_name*_i where i is the index of the unique conformer.  If not specified, the base name becomes simply *conformer*
        name : an optional name for the calculation directory.
        mol_info (dict) : an optional dictionary containing information will be written to the Compound Data section within the COSKF file.

    """

    _result_type = ADFCOSMORSConfResults

    def __init__(
        self,
        molecule: Molecule,
        conf_gen=None,
        first_filter=None,
        additional=None,
        final_filter=None,
        adf_singlepoint=False,
        initial_conformers=500,
        coskf_dir=None,
        coskf_name=None,
        mol_info={},
        **kwargs,
    ):

        super().__init__(children={}, **kwargs)

        self.job_count = 0

        self.mol = molecule
        mol_info["Molar Mass"] = molecule.get_mass()
        mol_info["Formula"] = molecule.get_formula()
        try:
            rings = molecule.locate_rings()
            flatten_atoms = [atom for subring in rings for atom in subring]
            nring = len(set(flatten_atoms))
            mol_info["Nring"] = int(nring)
        except:
            pass
        self.mol_info = mol_info

        self.adf_results = False
        self.cosmo_results = False

        self.adf_singlepoint = adf_singlepoint

        self.coskf_dir = coskf_dir
        self.coskf_name = coskf_name

        if self.coskf_dir is not None and not os.path.exists(self.coskf_dir):
            os.mkdir(self.coskf_dir)

        self.initial_conformers = initial_conformers

        if conf_gen is None:
            self.conf_gen = self.default_confgen()
        else:
            if not isinstance(conf_gen, ConformersJob):
                print(
                    "Wrong type for argument conf_gen.  Expected ConformersJob instance.  Using the default conformer generator."
                )
                self.conf_gen = self.default_confgen()
            else:
                self.conf_gen = conf_gen

        self.job_settings = [self.conf_gen.settings]
        self.filters = [None, first_filter]

        if additional is not None:
            for js, f in additional:
                self.job_settings.append(js)
                self.filters.append(f)

        self.final_filter = final_filter
        self.filters.append(final_filter)

        if not self.has_valid_filter_settings():
            pass

        self.children["job_0"] = self.conf_gen

    def default_confgen(self):

        sett = Settings()
        sett.input.AMS.Generator.RDKit
        sett.input.AMS.Generator.RDKit.InitialNConformers = self.initial_conformers
        return ConformersJob(name="conformers_uff", molecule=self.mol, settings=sett)

    def make_intermediate_job(self, settings):

        settings.input.AMS.InputConformersSet = self.children[f"job_{self.job_count-1}"].results
        self._add_filter(settings)
        return ConformersJob(name=f"additional_{self.job_count}", settings=settings)

    def make_adf_job(self):

        sett = ADFCOSMORSCompoundJob.adf_settings(False)
        if not self.adf_singlepoint:
            sett.input.AMS.Task = "Optimize"
            sett.input.AMS.GeometryOptimization.UseAMSWorker = "False"
            # sett.input.ams.GeometryOptimization.ConvergenceQuality = 'Good'
        else:
            sett.input.AMS.Task = "Score"
        sett.input.AMS.InputConformersSet = self.children[f"job_{self.job_count-1}"].results
        self._add_filter(sett)

        return ConformersJob(name="adf_conformers", settings=sett)

    def make_filter_job(self):
        sett = Settings()
        sett.input.AMS.Task = "Filter"
        sett.input.AMS.InputConformersSet = self.children["adf_job"].results
        self._add_filter(sett)

        return ConformersJob(name="adf_filter", settings=sett)

    def make_cosmo_job(self):

        sett = ADFCOSMORSCompoundJob.adf_settings(True, elements=list(set(at.symbol for at in self.mol)))
        sett.input.AMS.Task = "Replay"

        if self.final_filter is not None:
            previous_job = "filter_job"
        else:
            previous_job = "adf_job"

        self.children[previous_job].results.wait()
        sett.input.AMS.Replay.File = self.children[previous_job].results["conformers.rkf"]
        sett.input.AMS.Replay.StoreAllResultFiles = "True"

        return AMSJob(name="replay", settings=sett)

    def new_children(self):
        """Don't doc this

        :meta private:
        """
        self.job_count += 1
        if self.job_count < len(self.job_settings):
            settings = self.job_settings[self.job_count]
            new_job = self.make_intermediate_job(settings)
            return {f"job_{self.job_count}": new_job}

        if "adf_job" not in self.children:
            return {"adf_job": self.make_adf_job()}

        if "filter_job" not in self.children and self.final_filter is not None:
            return {"filter_job": self.make_filter_job()}

        if "cosmo_job" not in self.children:
            return {"cosmo_job": self.make_cosmo_job()}

        return None

    def postrun(self):
        self._make_coskfs()

    def _make_coskfs(self):

        base_name = self.coskf_name if self.coskf_name is not None else "conformer"
        if self.final_filter is not None:
            previous_job = "filter_job"
        else:
            previous_job = "adf_job"

        if self.coskf_dir is None:
            self.coskf_dir = self.children[previous_job].path
        for i, E in enumerate(self.children[previous_job].results.get_energies("Ha")):
            if f"Frame{i+1}.rkf" in self.children["cosmo_job"].results:
                cosmo_section = self.children["cosmo_job"].results.read_rkf_section("COSMO", f"Frame{i+1}")
                cosmo_section["Gas Phase Bond Energy"] = E
                name = f"{base_name}_{i}.coskf"

                fullname = os.path.join(self.coskf_dir, name)
                if os.path.exists(fullname):
                    os.remove(fullname)

                coskf = KFFile(os.path.join(self.coskf_dir, name), autosave=False)
                for key, val in cosmo_section.items():
                    coskf.write("COSMO", key, val)

                for key, value in self.mol_info.items():
                    # print(f"write to coskf {key}: {value}")
                    coskf.write("Compound Data", key, value)
                coskf.save()

    def _add_filter(self, sett):

        filt = self.filters[self.job_count]
        if filt is not None:
            if filt.max_num_confs is not None:
                sett.input.AMS.InputMaxConfs = filt.max_num_confs
            if filt.max_energy_range is not None:
                sett.input.AMS.InputMaxEnergy = filt.max_energy_range

    def has_valid_filter_settings(self):

        for js, f in zip(self.job_settings, self.filters):
            if js is not None and not isinstance(js, Settings):
                return False
            if f is not None and not isinstance(f, ADFCOSMORSConfFilter):
                return False
        return True



import os
from collections import OrderedDict
from typing import List, Optional, Union, Dict, Mapping

from scm.plams.interfaces.adfsuite.ams import AMSJob
from scm.plams.interfaces.adfsuite.crs import CRSJob
from scm.plams.tools.kftools import KFFile
from scm.plams.tools.periodic_table import PeriodicTable
from scm.plams.mol.molecule import Molecule
from scm.plams.core.basejob import MultiJob
from scm.plams.core.results import Results
from scm.plams.core.settings import Settings
from scm.plams.core.functions import add_to_instance
from scm.plams.interfaces.adfsuite.quickjobs import model_to_settings

__all__ = ["ADFCOSMORSCompoundJob", "ADFCOSMORSCompoundResults"]


class ADFCOSMORSCompoundResults(Results):
    """Results class for ADFCOSMORSCompoundJob"""

    def coskfpath(self):
        """
        Returns the path to the resulting .coskf
        """

        return os.path.join(self.job.path, self.job.name + ".coskf")

    def get_main_molecule(self):
        """
        Returns the optimized molecule
        """

        return self.job.children["solv"].results.get_main_molecule()

    def get_input_molecule(self):
        """
        Returns the input molecule
        """
        for job in self.job.children.values():
            return job.results.get_input_molecule()

    def get_sigma_profile(self, subsection: str = "profil"):
        """
        Returns the sigma profile of the molecule. For more details see `CRSResults.get_sigma_profile`.
        """
        return self.job.children["crs"].results.get_sigma_profile(subsection=subsection)


class ADFCOSMORSCompoundJob(MultiJob):
    """
    A class for performing the equivalent of Task: COSMO-RS Compound in the AMS GUI

    Args:
        molecule : a PLAMS |Molecule|

    Keyword Args:
        coskf_name  : A name for the generated .coskf file.  If nothing is specified, the name of the job will be used.
        coskf_dir  : The directory in which to place the generated .coskf file.  If nothing is specified, the file will be put in the plams directory corresponding to the job.
        preoptimization  : If None, do not preoptimize with a fast engine (then initial optimization is done with ADF). Otherwise, can be one of 'UFF', 'GAFF', 'GFNFF', 'GFN1-xTB', 'ANI-2x'. Note that you need valid licenses for ForceField or DFTB or MLPotential to use these preoptimizers.
        singlepoint (bool) :  Run a singlepoint in gasphase and with solvation to generate the .coskf file on the given Molecule. (no geometry optimization). Cannot be combined with ``preoptimization``.
        settings (Settings) : A |Settings| object.  settings.runscript.nproc, settings.input.adf.custom_options. If 'adf' is in settings.input it should be provided without the solvation block.
        name : an optional name for the calculation directory
        mol_info (dict) : an optional dictionary containing information will be written to the Compound Data section within the COSKF file.

    Example:

        .. code-block:: python

            mol = from_smiles('O')
            job = ADFCOSMORSCompoundJob(
                molecule = mol,
                preoptimization = 'UFF',
                coskf_dir = 'coskfs',
                coskf_name = 'Water',
                name = 'H2O',
                mol_info = {'CAS':'7732-18-5'}
            )
            job.run()
            print(job.results.coskfpath())

    """

    _result_type = ADFCOSMORSCompoundResults

    def __init__(
        self,
        molecule: Union[Molecule, None],
        coskf_name: Optional[str] = None,
        coskf_dir: Optional[str] = None,
        preoptimization: Optional[str] = None,
        singlepoint: bool = False,
        settings: Optional[Settings] = None,
        mol_info: Optional[Mapping[str, Union[float, int, str]]] = None,
        **kwargs,
    ):
        """

        Class for running the equivalent of "COSMO-RS Compound" in the AMS
        GUI. Note that these are ADF calculations, not COSMO-RS
        calculations!

        Initialize two or three jobs:

        (optional): Preoptimization with force field or semi-empirical method
        1. Gasphase optimization (BP86, DZP)
        2. Gasphase optimization (BP86, TZP, BeckeGrid Quality Good)
        3. Take optimized structure and run singlepoint with implicit solvation

        Access the result .coskf file with ``job.results.coskfpath()``.
        Note: this file will be called jobname.coskf, where jobname is the
        name of the ADFCOSMORSCompoundJob.

        """
        if preoptimization and singlepoint:
            raise ValueError("Cannot combine preoptimization with singlepoint")

        MultiJob.__init__(self, children=OrderedDict(), **kwargs)
        self.input_molecule = molecule

        self.mol_info = dict()
        if mol_info is not None:
            self.mol_info.update(mol_info)
        # self.mol_info = mol_info
        self.atomic_ion = False  # should be set when molecule is set if using a custom prerun() method
        if molecule is not None:
            self.mol_info["Molar Mass"] = molecule.get_mass()
            self.mol_info["Formula"] = molecule.get_formula()
            self.atomic_ion = len(molecule.atoms) == 1
            try:
                rings = molecule.locate_rings()
                flatten_atoms = [atom for subring in rings for atom in subring]
                nring = len(set(flatten_atoms))
                self.mol_info["Nring"] = int(nring)
            except:
                pass

        self.settings = settings or Settings()

        self.coskf_name = coskf_name
        self.coskf_dir = coskf_dir

        if self.coskf_dir is not None and not os.path.exists(self.coskf_dir):
            os.mkdir(self.coskf_dir)

        if self.coskf_name is not None and isinstance(self.coskf_name, str) and not self.coskf_name.endswith(".coskf"):
            self.coskf_name += ".coskf"

        gas_s = Settings()
        gas_s += self.adf_settings(solvation=False, settings=self.settings)
        gas_job = AMSJob(settings=gas_s, name="gas")

        if not singlepoint:
            if preoptimization:
                preoptimization_s = Settings()
                preoptimization_s.runscript.nproc = 1
                preoptimization_s.input.ams.Task = "GeometryOptimization"
                preoptimization_s += model_to_settings(preoptimization)
                preoptimization_job = AMSJob(
                    settings=preoptimization_s, name="preoptimization", molecule=self.input_molecule
                )
                self.children["preoptimization"] = preoptimization_job

            gas_s = Settings()
            gas_s.input.ams.Task = "GeometryOptimization"
            gas_s += self.adf_settings(solvation=False, settings=self.settings)
            gas_job = AMSJob(settings=gas_s, name="gas")

            if preoptimization:

                @add_to_instance(gas_job)
                def prerun(self):  # noqa: F811
                    self.molecule = self.parent.children["preoptimization"].results.get_main_molecule()

            else:
                gas_job.molecule = self.input_molecule

            self.children["gas"] = gas_job

        solv_s = Settings()
        solv_s.input.ams.Task = "SinglePoint"
        solv_job = AMSJob(settings=solv_s, name="solv")

        if singlepoint:

            gas_job.settings.input.ams.Task = "SinglePoint"

            @add_to_instance(gas_job)
            def prerun(self):  # noqa: F811
                self.molecule = self.parent.input_molecule

            self.children["gas"] = gas_job

        @add_to_instance(solv_job)
        def prerun(self):  # noqa: F811
            gas_job.results.wait()
            self.settings.input.ams.EngineRestart = "../gas/adf.rkf"
            self.settings.input.ams.LoadSystem.File = "../gas/ams.rkf"
            molecule_charge = gas_job.results.get_main_molecule().properties.get("charge", 0)
            self.settings.input.ams.LoadSystem._1 = f"# {self.parent.name}"
            self.settings.input.ams.LoadSystem._2 = f"# charge {molecule_charge}"
            self.settings += self.parent.adf_settings(
                solvation=True,
                settings=self.parent.settings,
                elements=list(set(at.symbol for at in self.parent.input_molecule)),
                atomic_ion=self.parent.atomic_ion,
            )
            # self.settings.input.ams.EngineRestart = self.parent.children['gas'].results.rkfpath(file='adf') # this doesn't work with PLAMS restart since the file will refer to the .res directory (so the job is rerun needlessly)
            # self.settings.input.ams.LoadSystem.File = self.parent.children['gas'].results.rkfpath(file='ams')
            # cannot copy to gasphase-ams.rkf etc. because that conflicts with PLAMS restarts
            # shutil.copyfile(gas_job.results.rkfpath(file='ams'), os.path.join(self.path, 'gasphase-ams.rkf'))
            # shutil.copyfile(gas_job.results.rkfpath(file='adf'), os.path.join(self.path, 'gasphase-adf.rkf'))

        @add_to_instance(solv_job)
        def postrun(self):
            self.parent.convert_to_coskf(
                self.results.rkfpath(file="adf"),
                os.path.join(
                    self.parent.coskf_dir if self.parent.coskf_dir is not None else self.parent.path,
                    self.parent.coskf_name if self.parent.coskf_name is not None else self.parent.name + ".coskf",
                ),
                self.parent.mol_info,
            )

        self.children["solv"] = solv_job

        sigma_s = Settings()
        sigma_s.input.property._h = "PURESIGMAPROFILE"

        compounds = [Settings()]
        sigma_s.input.compound = compounds
        crsjob = CRSJob(settings=sigma_s, name="sigma")

        @add_to_instance(crsjob)
        def prerun(self):  # noqa F811
            self.parent.children["solv"].results.wait()
            self.settings.input.compound[0]._h = os.path.join(
                self.parent.path if self.parent.coskf_dir is None else os.path.join(os.getcwd(), self.parent.coskf_dir),
                self.parent.coskf_name if self.parent.coskf_name is not None else self.parent.name + ".coskf",
            )

        self.children["crs"] = crsjob

    @staticmethod
    def _get_radii() -> Dict[str, float]:
        """Method to get the atomic radii from solvent.txt (for some elements the radii are instead the Klamt radii)"""
        with open(os.path.expandvars("$AMSHOME/data/gui/solvent.txt"), "r") as f:
            mod_allinger_radii = [float(x) for i, x in enumerate(f) if i > 0]
        radii = {PeriodicTable.get_symbol(i): r for i, r in enumerate(mod_allinger_radii, 1) if i <= 118}
        klamt_radii = {
            "H": 1.30,
            "C": 2.00,
            "N": 1.83,
            "O": 1.72,
            "F": 1.72,
            "Si": 2.48,
            "P": 2.13,
            "S": 2.16,
            "Cl": 2.05,
            "Br": 2.16,
            "I": 2.32,
        }
        radii.update(klamt_radii)

        return radii

    @staticmethod
    def solvation_settings(elements: Optional[List[str]] = None, atomic_ion: bool = False) -> Settings:
        sett = Settings()

        radii = {
            "H": 1.3,
            "He": 1.275,
            "Li": 2.125,
            "Be": 1.858,
            "B": 1.792,
            "C": 2.0,
            "N": 1.83,
            "O": 1.72,
            "F": 1.72,
            "Ne": 1.333,
            "Na": 2.25,
            "Mg": 2.025,
            "Al": 1.967,
            "Si": 2.48,
            "P": 2.13,
            "S": 2.16,
            "Cl": 2.05,
            "Ar": 1.658,
            "K": 2.575,
            "Ca": 2.342,
            "Sc": 2.175,
            "Ti": 1.992,
            "V": 1.908,
            "Cr": 1.875,
            "Mn": 1.867,
            "Fe": 1.858,
            "Co": 1.858,
            "Ni": 1.85,
            "Cu": 1.883,
            "Zn": 1.908,
            "Ga": 2.05,
            "Ge": 2.033,
            "As": 1.967,
            "Se": 1.908,
            "Br": 2.16,
            "Kr": 1.792,
            "Rb": 2.708,
            "Sr": 2.5,
            "Y": 2.258,
            "Zr": 2.117,
            "Nb": 2.025,
            "Mo": 1.992,
            "Tc": 1.967,
            "Ru": 1.95,
            "Rh": 1.95,
            "Pd": 1.975,
            "Ag": 2.025,
            "Cd": 2.083,
            "In": 2.2,
            "Sn": 2.158,
            "Sb": 2.1,
            "Te": 2.033,
            "I": 2.32,
            "Xe": 1.9,
            "Cs": 2.867,
            "Ba": 2.558,
            "La": 2.317,
            "Ce": 2.283,
            "Pr": 2.275,
            "Nd": 2.275,
            "Pm": 2.267,
            "Sm": 2.258,
            "Eu": 2.45,
            "Gd": 2.258,
            "Tb": 2.25,
            "Dy": 2.242,
            "Ho": 2.225,
            "Er": 2.225,
            "Tm": 2.225,
            "Yb": 2.325,
            "Lu": 2.208,
            "Hf": 2.108,
            "Ta": 2.025,
            "W": 1.992,
            "Re": 1.975,
            "Os": 1.958,
            "Ir": 1.967,
            "Pt": 1.992,
            "Au": 2.025,
            "Hg": 2.108,
            "Tl": 2.158,
            "Pb": 2.283,
            "Bi": 2.217,
            "Po": 2.158,
            "At": 2.092,
            "Rn": 2.025,
            "Fr": 3.033,
            "Ra": 2.725,
            "Ac": 2.567,
            "Th": 2.283,
            "Pa": 2.2,
            "U": 2.1,
            "Np": 2.1,
            "Pu": 2.1,
            "Am": 2.1,
            "Cm": 2.1,
            "Bk": 2.1,
            "Cf": 2.1,
            "Es": 2.1,
            "Fm": 2.1,
            "Md": 2.1,
            "No": 2.1,
            "Lr": 2.1,
            "Rf": 2.1,
            "Db": 2.1,
            "Sg": 2.1,
            "Bh": 2.1,
            "Hs": 2.1,
            "Mt": 2.1,
            "Ds": 2.1,
            "Rg": 2.1,
            "Cn": 2.1,
            "Nh": 2.1,
            "Fl": 2.1,
            "Mc": 2.1,
            "Lv": 2.1,
            "Ts": 2.1,
            "Og": 2.1,
        }  # from _get_radii()

        if elements:
            radii = {k: radii[k] for k in sorted(elements)}

        if atomic_ion:
            charge_method = "method=atom corr"
        else:
            charge_method = "method=Conj corr"

        sett.input.adf.solvation = {
            "surf": "Delley",
            "solv": "name=CRS cav0=0.0 cav1=0.0",
            "charged": charge_method,
            "c-mat": "Exact",
            "scf": "Var All",
            "radii": radii,
        }
        return sett

    @staticmethod
    def adf_settings(
        solvation: bool, settings=None, elements: Optional[List[str]] = None, atomic_ion: bool = False
    ) -> Settings:
        """
        Returns ADF settings with or without solvation

        If solvation == True, then also include the solvation block.
        """

        s = Settings()
        if settings:
            s = settings.copy()
        if "basis" not in s.input.adf and "xc" not in s.input.adf:
            s.input.adf.Basis.Type = "TZP"
            s.input.adf.Basis.Core = "Small"
            s.input.adf.XC.GGA = "BP86"
            s.input.adf.Symmetry = "NOSYM"
            s.input.adf.BeckeGrid.Quality = "Good"
        if solvation:
            s += ADFCOSMORSCompoundJob.solvation_settings(elements=elements, atomic_ion=atomic_ion)
        return s

    @staticmethod
    def convert_to_coskf(rkf: str, coskf: str, mol_info: Optional[Mapping[str, Union[float, int, str]]] = None):
        """
        rkf: str
            absolute path to adf.rkf

        coskf: str
            path to write out the resulting .coskf file

        mol_info: dict[str, float | int | str], optional
            Optional information to write out in the "Compound Data" section of the .coskf file

        """
        f = KFFile(rkf)
        cosmo = f.read_section("COSMO")
        coskf_file = KFFile(coskf, autosave=False)
        for k, v in cosmo.items():
            coskf_file.write("COSMO", k, v)
        if mol_info is not None:
            for key, value in mol_info.items():
                # print(f"write to coskf {key}: {value}")
                coskf_file.write("Compound Data", key, value)
        coskf_file.save()



from os.path import join as opj

from scm.plams.core.basejob import SingleJob
from scm.plams.core.errors import PlamsError
from scm.plams.core.results import Results
from scm.plams.core.settings import Settings
from scm.plams.mol.molecule import Molecule
from scm.plams.tools.units import Units

__all__ = ["VASPJob", "VASPResults"]


class VASPResults(Results):
    """A class for VASP results."""

    def get_energy(self, index=-1, unit="a.u."):
        """Returns sigma->0 (!!!) energy without entropy."""
        s = self.grep_output("energy  without entropy=")[index]
        if not isinstance(index, slice):
            return Units.convert(float(s.split()[-1]), "eV", unit)
        else:
            return [Units.convert(float(x.split()[-1]), "eV", unit) for x in s]

    def get_dispersion_energy(self, index=-1, unit="a.u."):
        """Returns `Edisp (eV)` from the OUTCAR."""
        s = self.grep_output("Edisp (eV)")[index]
        if not isinstance(index, slice):
            return Units.convert(float(s.split()[-1]), "eV", unit)
        else:
            return [Units.convert(float(x.split()[-1]), "eV", unit) for x in s]


class VASPJob(SingleJob):
    """
    A class representing a single computational job with `VASP <https://www.vasp.at/>`

    * Set 'ignore_molecule' in ``self.settings`` to disable Molecule handling through ASE.
    * Set 'ignore_potcar' in ``self.settings`` to disable automatic `POTCAR` creation.
    * Set the path to the `POTCAR` files in  ``self.settings.input.potcar`` for automatic `POTCAR` creation.
    * If `POTCAR` files not matching the element symbol should be used, give a translation dict in ``self.settings.input.potcardict``.
        E.g. `{'Fe': 'Fe_pv'}`.
    * Settings branch ``input.incar`` is parsed into the `INCAR` file, ``input.xxx`` into the corresponding `XXX` file.
    * Use the PLAMS notation `_h`, `_1`, `_2`, ... to obtain keywords in specific order (e.g. for the KPOINTS file).
    """

    _command = "vasp_std"
    _filenames = {"inp": "INCAR", "run": "$JN.run", "out": "OUTCAR", "err": "$JN.err", "log": "$JN.log"}
    _result_type = VASPResults

    def get_input(self):
        """
        Transform all contents of ``input`` branch of ``settings`` into string.
        """

        def vaspstr(inp):
            # convert to VASP type string
            if isinstance(inp, Settings):
                raise PlamsError("Nested Settings Object not supported in VASP parser.")
            if isinstance(inp, bool):
                if inp:
                    return ".TRUE."
                else:
                    return ".FALSE."
            elif isinstance(inp, list):
                return " ".join([str(x) for x in inp])
            else:
                return str(inp).upper()

        def parse(key, value):
            ret = ""

            if isinstance(value, Settings):
                if "_h" in value:
                    ret += "{}\n".format(vaspstr(value["_h"]))
                i = 1
                while ("_" + str(i)) in value:
                    ret += "{}\n".format(vaspstr(value["_" + str(i)]))
                    i += 1
                for el in value:
                    if not el.startswith("_"):
                        ret += parse(el, value[el])

            elif isinstance(value, list):
                ret += "{}\n".format(vaspstr(value))
            else:
                ret += "{} = {}\n".format(key.upper(), vaspstr(value).upper())
            return ret

        use_molecule = ("ignore_molecule" not in self.settings) or (self.settings.ignore_molecule == False)
        use_potcar = ("ignore_potcar" not in self.settings) or (self.settings.ignore_potcar == False)
        if use_molecule:
            self._parsemol()
        if use_potcar:
            self._parsepotcar()

        # we need all first-level keys in uppercase
        tmp = Settings()
        for key in self.settings.input:
            tmp[key.upper()] = self.settings.input[key]

        for item in tmp:
            # POTCAR creation handled above
            if "POTCAR" in item:
                continue
            # INCAR
            elif item == "INCAR":
                inp = parse(item, tmp[item])
            else:
                with open(opj(self.path, item), "w") as f:
                    f.write(parse(item, tmp[item]))
        return inp

    def _parsemol(self):
        if "ase" in Molecule._writeformat:
            # ASE has a write function for VASP coordinate files, use that if possible
            filename = opj(self.path, "POSCAR")
            self.molecule.writease(filename)
        else:
            raise PlamsError("VASP Interface has no builtin Molecule support, install ASE. See Doc for details.")

    def _parsepotcar(self):
        tree = self.settings.input
        if "potcar" in tree:
            elements = [self.molecule.atoms[0].symbol]
            for atom in self.molecule.atoms[1:]:
                if not atom.symbol == elements[-1]:
                    elements.append(atom.symbol)
            if "potcardict" in tree:
                translate = dict(tree.potcardict)
            else:
                translate = {el: el for el in set(elements)}
            # open files
            files = [open(opj(tree.potcar, translate[el], "POTCAR"), "r") for el in elements]

            with open(opj(self.path, "POTCAR"), "w") as f:
                for potcar in files:
                    f.write(potcar.read())
                    potcar.close()
        else:
            raise PlamsError("VASP Interface needs the POTCAR path in self.settings.input.potcar.")

    def get_runscript(self):
        """
        Run VASP

        Overwrite ``self._command`` to change the default VASP Binary.
        """
        ret = self._command
        if self.settings.runscript.stdout_redirect:
            ret += " >" + self._filename("log")
        ret += "\n\n"
        return ret

    def check(self):
        """
        Look for the normal termination line in output. Note, that does not mean your calculation was successful!
        """
        termination = self.results.grep_output("General timing and")
        return bool(termination)



import pytest
import builtins
from importlib import reload
from threading import Thread

from scm.plams.core.settings import Settings
from scm.plams.unit_tests.test_helpers import get_mock_import_function, skip_if_no_ams_installation


@pytest.fixture
def ams_text_inputs():
    """
    Set of example AMS text inputs and the corresponding expected settings objects.
    """

    geometry_optimisation_input = """
Task GeometryOptimization

Engine ADF
  NumericalQuality Basic
  XC
    GGA PBE
  End
  basis
    core None
    type DZ
  End
EndEngine
"""

    geometry_optimisation_settings = Settings()
    geometry_optimisation_settings.adf.basis.core = "None"
    geometry_optimisation_settings.adf.basis.type = "DZ"
    geometry_optimisation_settings.adf.numericalquality = "Basic"
    geometry_optimisation_settings.adf.xc.gga = "PBE"
    geometry_optimisation_settings.ams.task = "GeometryOptimization"

    md_input = """
MolecularDynamics
  Checkpoint
    Frequency 1000000
  End
  InitialVelocities
    Temperature 600.0
  End
  NSteps 100000
  Shake
    All Bonds H *
  End
  Thermostat
    Tau 100
    Temperature 600.0
    Type Berendsen
  End
  TimeStep 1.0
  Trajectory
    SamplingFreq 100
  End
End
"""

    md_settings = Settings()
    md_thermo_settings = Settings()
    md_settings.ams.moleculardynamics.checkpoint.frequency = "1000000"
    md_settings.ams.moleculardynamics.initialvelocities.temperature = "600.0"
    md_settings.ams.moleculardynamics.nsteps = "100000"
    md_settings.ams.moleculardynamics.shake.all = ["Bonds H *"]
    md_thermo_settings.tau = "100"
    md_thermo_settings.temperature = "600.0"
    md_thermo_settings.type = "Berendsen"
    md_settings.ams.moleculardynamics.thermostat = [md_thermo_settings]
    md_settings.ams.moleculardynamics.timestep = "1.0"
    md_settings.ams.moleculardynamics.trajectory.samplingfreq = "100"

    yield [(geometry_optimisation_input, geometry_optimisation_settings), (md_input, md_settings)]

    # Tear-down: reload module without the patched import function
    import scm.plams.interfaces.adfsuite.inputparser as inputparser

    reload(inputparser)


@pytest.fixture
def system_text_inputs():
    """
    Set of example system input texts and the corresponding expected settings objects.
    """

    water_system_input = """
System
  Atoms
              O       0.0000000000       0.0000000000       0.0000000000
              H       1.0000000000       0.0000000000       0.0000000000
              H       0.0000000000       1.0000000000       0.0000000000
  End
End
    """

    water_settings = Settings()
    water_settings.atoms._1 = [
        "O       0.0000000000       0.0000000000       0.0000000000",
        "H       1.0000000000       0.0000000000       0.0000000000",
        "H       0.0000000000       1.0000000000       0.0000000000",
    ]
    water_system_settings = Settings()
    water_system_settings.system = [water_settings]

    graphene_system_input = """
System
   Atoms
      C   -1.2300000000000000 -2.1304224999999999  0.0000000000000000
      C   -1.2300000000000000 -0.7101408299999999  0.0000000000000000
      C    0.0000000000000000  0.0000000000000000  0.0000000000000000
      C    0.0000000000000000  1.4202816700000001  0.0000000000000000
   End
   Lattice
        2.4600000000000000   0.0000000000000000   0.0000000000000000
        0.0000000000000000   4.2608449999999998   0.0000000000000000
        0.0000000000000000   0.0000000000000000  80.0000000000000000
   End
End
"""
    graphene_settings = Settings()
    graphene_settings.atoms._1 = [
        "C   -1.2300000000000000 -2.1304224999999999  0.0000000000000000",
        "C   -1.2300000000000000 -0.7101408299999999  0.0000000000000000",
        "C    0.0000000000000000  0.0000000000000000  0.0000000000000000",
        "C    0.0000000000000000  1.4202816700000001  0.0000000000000000",
    ]
    graphene_settings.lattice._1 = [
        "2.4600000000000000   0.0000000000000000   0.0000000000000000",
        "0.0000000000000000   4.2608449999999998   0.0000000000000000",
        "0.0000000000000000   0.0000000000000000  80.0000000000000000",
    ]
    graphene_system_settings = Settings()
    graphene_system_settings.system = [graphene_settings]

    yield [(water_system_input, water_system_settings), (graphene_system_input, graphene_system_settings)]

    # Tear-down: reload module without the patched import function
    import scm.plams.interfaces.adfsuite.inputparser as inputparser

    reload(inputparser)


@pytest.fixture
def large_system_input():
    """
    Input with large molecule.
    """
    large_system_input = """
    System
       Atoms
    """
    for i in range(100):
        for j in range(100):
            large_system_input += f"""        C    {i}    {j}    0
"""
    large_system_input += """
  End
End
"""

    return large_system_input


def test_to_settings_without_scmlibbase_succeeds(ams_text_inputs, monkeypatch):
    input_parser = get_monkeypatched_input_parser(monkeypatch)
    to_settings_succeeds(ams_text_inputs, input_parser)


def test_to_settings_with_scmlibbase_succeeds(ams_text_inputs):
    input_parser = get_input_parser_or_skip()
    to_settings_succeeds(ams_text_inputs, input_parser)


def test_to_dict_without_scmlibbase_succeeds(system_text_inputs, monkeypatch):
    input_parser = get_monkeypatched_input_parser(monkeypatch)
    to_dict_succeeds(system_text_inputs, input_parser)


def test_to_dict_with_scmlibbase_succeeds(system_text_inputs):
    input_parser = get_input_parser_or_skip()
    to_dict_succeeds(system_text_inputs, input_parser)


def get_monkeypatched_input_parser(monkeypatch):
    # If there is no AMS installation the input parser will not run so skip test with a warning
    skip_if_no_ams_installation()

    # Mock scm.libbase import failing (even when present in the env)
    mock_import_function = get_mock_import_function("scm.libbase")
    monkeypatch.setattr(builtins, "__import__", mock_import_function)

    # Reload the module without scm.libbase
    import scm.plams.interfaces.adfsuite.inputparser as inputparser

    reload(inputparser)
    from scm.plams.interfaces.adfsuite.inputparser import InputParserFacade, InputParser

    # Get an instance of the input parser facade using the fallback input parser
    input_parser = InputParserFacade()
    assert isinstance(input_parser.parser, InputParser)

    return input_parser


def get_input_parser_or_skip():
    # If there is no AMS installation the input parser will not run so skip test with a warning
    skip_if_no_ams_installation()

    from scm.plams.interfaces.adfsuite.inputparser import InputParserFacade, InputParser

    # Get an instance of the input parser facade using the scm.libbase parser
    # otherwise skip the test if the package is not loaded
    input_parser = InputParserFacade()
    if input_parser._has_scm_libbase:
        from scm.libbase import InputParser as InputParserScmLibbase

        assert isinstance(input_parser.parser, InputParserScmLibbase)
        return input_parser
    else:
        assert isinstance(input_parser.parser, InputParser)
        pytest.skip("Skipping test because optional 'scm.libbase' package is not available")


def to_settings_succeeds(input_texts, input_parser):
    for input_text, expected_settings in input_texts:
        actual_settings = input_parser.to_settings("ams", input_text)
        assert actual_settings == expected_settings


def to_dict_succeeds(input_texts, input_parser):
    for input_text, expected_settings in input_texts:
        actual_dict = input_parser.to_dict("ams", input_text)
        assert actual_dict == expected_settings.as_dict()


def test_libbase_parser_cached(ams_text_inputs):
    input_parser = get_input_parser_or_skip()

    assert "ams" not in input_parser.parser.parsers.keys()

    input_parser.to_dict("ams", ams_text_inputs[0][0])

    assert "ams" in input_parser.parser.parsers.keys()

    assert input_parser.parser == input_parser.parser


def test_thread_safe(large_system_input):
    get_input_parser_or_skip()

    errors = []

    def parse():
        try:
            input_parser = get_input_parser_or_skip()
            input_parser.to_settings("ams", large_system_input)
        except Exception as e:
            errors.append(e)

    threads = [Thread(target=parse) for _ in range(100)]
    for t in threads:
        t.start()
    for t in threads:
        t.join()

    assert len(errors) == 0



from scm.plams.interfaces.molecule.rdkit import writepdb
from scm.plams.mol.molecule import Molecule
from scm.plams.tools.periodic_table import PeriodicTable
from scm.plams.trajectories.rkfhistoryfile import molecules_to_rkf

try:
    from scm.plams.interfaces.molecule.ase import toASE
except ImportError:
    pass
import os
from typing import List, Tuple

from scm.plams.recipes.md.amsmdjob import AMSMDJob

"""


job = AMSMovingRestraintRMSDJob(
    visit_molecules=[mol1, mol2, mol3],
    kappa_scheduler=KappaExponentialScheduler(kappa=10000, n=4, base=10),
    nsteps=100000,
    ...
)

job = AMSMovingRestraintInternalJob(
    restraints=[
        BondRestraint(atom1, atom2, start=x, end=y, kappa_scheduler=KappaLinearScheduler(10000)),
        BondRestraint(atom1, atom2, start=x, end=y, kappa_scheduler=KappaLinearScheduler(10000)),
    ],
    consecutive=False
)

"""

__all__ = ["AMSMovingRestraintRMSDJob", "AMSMovingRestraintBondSwitchJob"]


def tolist(x):
    if not isinstance(x, list):
        x = [x]
    return x


def set_plumed_settings(settings, plumed_sett_list):
    """
    settings: Settings
        will be modified!

    plumed_sett_list: list of str
        The input to plumed. Should not include the final End
    """
    ret = ""
    indent = 6 * " "
    for line in plumed_sett_list:
        ret += f"{indent}{line}\n"
    ret += "    End"
    settings.input.ams.MolecularDynamics.Plumed.Input = ret


class AMSMovingRestraintBondSwitchJob(AMSMDJob):
    def __init__(
        self,
        visit_molecules: List[Molecule],
        kappa: List[float],
        start_step: int = 1,
        kappa_others: List[float] = None,
        **kwargs,
    ):
        AMSMDJob.__init__(self, **kwargs)
        self.visit_molecules = tolist(visit_molecules).copy()
        self.kappa = tolist(kappa)
        self.start_step = start_step
        self.nsteps = int(self.settings.input.ams.MolecularDynamics.NSteps)
        self.kappa_others = tolist(kappa_others) if kappa_others is not None else self.kappa

        previous_mol = self.molecule.copy()
        assert previous_mol is not None, "You must specify a molecule when initializing AMSMovingRestraintBondSwitchJob"

        if not previous_mol.bonds:
            previous_mol.guess_bonds()

        plumed_sett_list = [""]
        # plumed_sett_list.append('COORDINATION GROUPA=14 GROUPB=8,9,10,11,12,13,14,15 R_0=0.13 LABEL=c_14')
        # plumed_sett_list.append('PRINT ARG=c_14 FILE=c_14.txt STRIDE=1')
        counter = 0
        current_step = self.start_step
        nsteps_part = (self.nsteps - self.start_step) // len(self.visit_molecules)
        nsteps_per_kappa = nsteps_part // len(self.kappa)
        nsteps_per_kappa_others = nsteps_part // len(self.kappa_others)

        for counter, mol in enumerate(self.visit_molecules, 1):
            if not mol.bonds:
                mol.guess_bonds()

            bond_switches, other, nonbonded = self._get_bond_switches(previous_mol, mol)

            for oth in other:
                i1, i2, distance = oth
                label = f"bond{counter}_{i1}_{i2}"

                plumed_sett_list.append(f"DISTANCE ATOMS={i1},{i2} LABEL={label}")
                # now loop through all kappa values

                s = f"MOVINGRESTRAINT ARG={label}"
                for i, kap in enumerate(self.kappa_others, 0):
                    s += f" STEP{i}={current_step+nsteps_per_kappa_others*i} KAPPA{i}={kap} AT{i}={distance*0.1:.4f}"
                end_index = len(self.kappa_others)
                s += f" STEP{end_index}={current_step+nsteps_per_kappa_others*end_index} KAPPA{end_index}=0.0 AT{end_index}={distance*0.1:.4f}"

                plumed_sett_list.append(s)

            for bs in bond_switches:
                i1, i2, distance = bs
                label = f"bond{counter}_{i1}_{i2}"

                plumed_sett_list.append(f"DISTANCE ATOMS={i1},{i2} LABEL={label}")
                # now loop through all kappa values

                s = f"MOVINGRESTRAINT ARG={label}"
                for i, kap in enumerate(self.kappa, 0):
                    s += f" STEP{i}={current_step+nsteps_per_kappa*i} KAPPA{i}={kap} AT{i}={distance*0.1:.4f}"
                end_index = len(self.kappa)
                end_step = current_step + nsteps_per_kappa * end_index
                s += f" STEP{end_index}={end_step} KAPPA{end_index}=0.0 AT{end_index}={distance*0.1:.4f}"

                plumed_sett_list.append(s)

            for nb in nonbonded:
                i1, i2, distance = nb
                label = f"nonbond{counter}_{i1}_{i2}"

                plumed_sett_list.append(f"DISTANCE ATOMS={i1},{i2} LABEL={label}")
                # now loop through all kappa values

                s = f"MOVINGRESTRAINT ARG={label} VERSE=L"
                s += f" STEP0={current_step} KAPPA0=0 AT0={distance*0.1:.4f}"
                s += f" STEP1={end_step-1} KAPPA1=100000.0 AT1={distance*0.1:.4f}"
                s += f" STEP2={end_step} KAPPA2=0.0 AT2={distance*0.1:.4f}"

                plumed_sett_list.append(s)

            current_step += nsteps_per_kappa * end_index + 1

            previous_mol = mol

        set_plumed_settings(self.settings, plumed_sett_list)

    @staticmethod
    def _get_bond_switches(mol1: Molecule, mol2: Molecule) -> Tuple[List[Tuple[int, int, float]]]:
        assert len(mol1) == len(mol2), "mol1 and mol2 have different lengths!"
        bonds1 = set(tuple(sorted((mol1.index(b.atom1), mol1.index(b.atom2)))) for b in mol1.bonds)
        bonds2 = set(tuple(sorted((mol2.index(b.atom1), mol2.index(b.atom2)))) for b in mol2.bonds)

        all_reactive_atoms = set()

        atoms2 = toASE(mol2)

        # changed bonds
        targets = []
        t1 = bonds1 - bonds2
        t2 = bonds2 - bonds1
        for t in t1:
            d = atoms2.get_distance(t[0] - 1, t[1] - 1, mic=True)
            targets.append((t[0], t[1], d))
            all_reactive_atoms.add(t[0])
            all_reactive_atoms.add(t[1])
        for t in t2:
            d = atoms2.get_distance(t[0] - 1, t[1] - 1, mic=True)
            targets.append((t[0], t[1], d))
            all_reactive_atoms.add(t[0])
            all_reactive_atoms.add(t[1])

        all_non_reactive_atoms = set(range(1, len(mol1) + 1)) - all_reactive_atoms

        union_bonds = bonds1.union(bonds2)

        # other bonded
        other = []
        other_bonds = union_bonds - t1 - t2
        for ob in other_bonds:
            i1, i2 = ob
            # other.append( (i1, i2, 0.5*(mol1[i1].distance_to(mol1[i2]) + mol2[i1].distance_to(mol2[i2]))) )
            d = atoms2.get_distance(i1 - 1, i2 - 1, mic=True)
            other.append((i1, i2, d))

        # other nonbonded
        nonbonded = []
        covalent_radius_multiplier = 1.5

        for ireac in sorted(all_reactive_atoms):
            for inonreac in sorted(all_non_reactive_atoms):
                if (ireac, inonreac) in union_bonds or (inonreac, ireac) in union_bonds:
                    continue
                at1 = mol2[ireac]
                at2 = mol2[inonreac]
                r = PeriodicTable.get_radius(at1.symbol) + PeriodicTable.get_radius(at2.symbol)
                nonbonded.append((ireac, inonreac, covalent_radius_multiplier * r))
        # for i, at1 in enumerate(mol2, 1):
        #    for j, at2 in enumerate(mol2, 1):
        #        if j >= i:
        #            continue
        #        if (i, j) in bonds1 or (i, j) in bonds2:
        #            continue
        #        r = PeriodicTable.get_radius(at1.symbol) + PeriodicTable.get_radius(at2.symbol)
        #        nonbonded.append( (i, j, covalent_radius_multiplier*r) )

        return targets, other, nonbonded


class AMSMovingRestraintRMSDJob(AMSMDJob):
    def __init__(self, visit_molecules: List[Molecule], kappa: List[int], start_step: int = 1, **kwargs):
        AMSMDJob.__init__(self, **kwargs)
        self.visit_molecules = tolist(visit_molecules)
        self.kappa = tolist(kappa)
        self.start_step = start_step
        self.nsteps = int(self.settings.input.ams.MolecularDynamics.NSteps)
        self.pdb_files = self._get_pdb_file_names()

    def _get_pdb_file_names(self):
        return [f"mol{i}.pdb" for i in range(len(self.visit_molecules))]

    def prerun(self):  # noqa F811
        plumed_sett_list = [""]
        current_step = self.start_step
        nsteps_part = (self.nsteps - self.start_step) // len(self.visit_molecules)
        nsteps_per_kappa = nsteps_part // len(self.kappa)

        molecules_to_rkf(self.visit_molecules, os.path.join(self.path, "to_be_visited.rkf"))

        for pdb_rel, mol in zip(self.pdb_files, self.visit_molecules):
            pdb = os.path.join(self.path, pdb_rel)
            self._write_plumed_compatible_pdb(pdb, mol)
            name = os.path.splitext(pdb_rel)[0]

            label = f"rmsd_{name}"

            indent = 6 * " "
            plumed_sett_list.append(f"{indent}RMSD REFERENCE={pdb} LABEL={label}")
            # now loop through all kappa values

            s = f"{indent}MOVINGRESTRAINT ARG={label}"
            for i, kap in enumerate(self.kappa, 0):
                s += f" STEP{i}={current_step+nsteps_per_kappa*i} KAPPA{i}={kap} AT{i}=0.0"
            end_index = len(self.kappa)
            s += f" STEP{end_index}={current_step+nsteps_per_kappa*end_index} KAPPA{end_index}=0.0 AT{end_index}=0.0"

            current_step += nsteps_per_kappa * end_index + 1

            plumed_sett_list.append(s)

        plumed_sett_list.append("    End")

        self.settings.input.ams.MolecularDynamics.Plumed.Input = "\n".join(plumed_sett_list)

    @classmethod
    def _write_plumed_compatible_pdb(cls, pdb_file: str, molecule: Molecule):
        temporary_pdb = "temp.pdb"
        writepdb(molecule, temporary_pdb)
        char_index = 62
        try:
            xyzfile = os.path.splitext(pdb_file)[0] + ".xyz"
            molecule.write(xyzfile)
            with open(pdb_file, "w") as f:
                with open(temporary_pdb, "r") as temp:
                    for line in temp:
                        splitline = list(line)
                        if len(splitline) > char_index + 1 and splitline[char_index] == "0":
                            splitline[char_index] = "1"
                        f.write("".join(splitline) + "\n")
        finally:
            if os.path.exists(temporary_pdb):
                os.remove(temporary_pdb)



#!/usr/bin/env plams
"""
Script for evaluating the HC7-11 and ISOL6 benchmark sets with different computational methods.

Modify the s_engine_dict variable below to specify which settings will be used.

A table with reaction energies will be printed to summary.txt
"""

summary_fname = "summary.txt"
with open(summary_fname, "w", buffering=1) as summary_file:
    summary_file.write(
        "#Method HC7_1 HC7_2 HC7_3 HC7_4 HC7_5 HC7_6 HC7_7 ISOL6_1 ISOL6_2 ISOL6_3 ISOL6_4 ISOL6_5 (kcal/mol)\n"
    )
    summary_file.write("Truhlar_ref 14.34 25.02 1.90 9.81 14.84 193.99 127.22 9.77 21.76 6.82 33.52 5.30\n")
    summary_file.write("Smith_wB97X_ref 28.77 41.09 1.75 6.26 9.30 238.83 157.65 9.32 20.80 1.03 26.43 0.40\n")

mol_dict = read_molecules("HC7-11")
mol_dict.update(read_molecules("ISOL6"))

s_reaxff = Settings()
s_reaxff.input.ReaxFF.ForceField = "CHON-2019.ff"

s_ani2x = Settings()
s_ani2x.input.MLPotential.Backend = "TorchANI"
s_ani2x.input.MLPotential.Model = "ANI-2x"

s_ani1ccx = Settings()
s_ani1ccx.input.MLPotential.Backend = "TorchANI"
s_ani1ccx.input.MLPotential.Model = "ANI-1ccx"

s_dftb = Settings()
s_dftb.input.DFTB.Model = "SCC-DFTB"
s_dftb.input.DFTB.ResourcesDir = "DFTB.org/mio-1-1"

s_band = Settings()
s_band.input.BAND.XC.libxc = "wb97x"
s_band.input.BAND.Basis.Type = "TZP"
s_band.input.BAND.Basis.Core = "None"

s_adf = Settings()
s_adf.input.ADF.Basis.Type = "TZP"
s_adf.input.ADF.Basis.Core = "None"
s_adf.input.ADF.XC.libxc = "WB97X"

s_mp2 = Settings()
s_mp2.input.ADF.Basis.Type = "TZ2P"
s_mp2.input.ADF.Basis.Core = "None"
s_mp2.input.ADF.XC.MP2 = ""

# The engines in s_engine_dict will be used for the calculation
# You can remove the engines that you would not like to run (for example, if you do not have the necessary license)
s_engine_dict = {
    "ANI-1ccx": s_ani1ccx,
    "ANI-2x": s_ani2x,
    "DFTB": s_dftb,
    "ReaxFF": s_reaxff,
    "ADF": s_adf,
    "MP2": s_mp2,
    "BAND": s_band,
}

s_ams = Settings()
s_ams.input.ams.Task = "SinglePoint"
# s_ams.input.ams.Task = 'GeometryOptimization'
# s_ams.input.ams.GeometryOptimization.CoordinateType = 'Cartesian'

jobs = dict()

with open(summary_fname, "a", buffering=1) as summary_file:
    for engine_name, s_engine in s_engine_dict.items():
        s = s_ams.copy() + s_engine.copy()
        jobs[engine_name] = dict()

        # call .run() for *all* jobs *before* accessing job.results.get_energy() for *any* job
        for mol_name, mol in mol_dict.items():
            jobs[engine_name][mol_name] = AMSJob(settings=s, molecule=mol, name=engine_name + "_" + mol_name)
            jobs[engine_name][mol_name].run()

    for engine_name in s_engine_dict:
        # for each engine, calculate reaction energies
        E = dict()
        for mol_name, job in jobs[engine_name].items():
            E[mol_name] = job.results.get_energy(unit="kcal/mol")
        deltaE_list = [
            ##### HC7/11 ######
            E["22"] - E["1"],
            E["31"] - E["1"],
            E["octane"] - E["2233tetramethylbutane"],
            5 * E["ethane"] - E["hexane"] - 4 * E["methane"],
            7 * E["ethane"] - E["octane"] - 6 * E["methane"],
            3 * E["ethylene"] + 2 * E["ethyne"] - E["adamantane"],
            3 * E["ethylene"] + 1 * E["ethyne"] - E["bicyclo222octane"],
            ###### start ISOL6 ######
            E["p_3"] - E["e_3"],
            E["p_9"] - E["e_9"],
            E["p_10"] - E["e_10"],
            E["p_13"] - E["e_13"],
            E["p_14"] - E["e_14"],
        ]

        out_str = engine_name
        for deltaE in deltaE_list:
            out_str += " {:.1f}".format(deltaE)
        out_str += "\n"

        print(out_str)
        summary_file.write(out_str)


# summary.txt will contain a table similar to the below
##Method          HC7_1   HC7_2  HC7_3  HC7_4  HC7_5   HC7_6   HC7_7 ISOL6_1 ISOL6_2 ISOL6_3 ISOL6_4 ISOL6_5 (kcal/mol)
# Truhlar_ref      14.34   25.02   1.90   9.81  14.84  193.99  127.22    9.77   21.76    6.82   33.52    5.30
# Smith_wB97X_ref  28.77   41.09   1.75   6.26   9.30  238.83  157.65    9.32   20.80    1.03   26.43    0.40
# ANI-1ccx         15.8    30.7   -0.2    7.7   11.7   196.5   127.9     9.2    21.5     3.8    35.4     6.9
# ANI-2x           42.4    48.1   -1.9    5.4    8.0   238.4   156.9     7.9    20.8    -1.5    26.7     0.5
# DFTB              7.1    19.0    0.2    3.6    5.4   223.5   150.8    11.7    22.6     7.9    35.2     8.6
# ReaxFF           34.3    23.0   13.4    2.6    5.5   289.2   168.3    -9.2    24.2    70.7    30.9    89.3
# ADF              20.8    31.4   -1.9    6.7    9.9   214.5   139.7     9.5    20.5     4.7    31.0     4.8
# MP2              20.5    27.1    5.4    9.9   15.0   214.8   141.3    11.0    24.6     8.4    36.8     6.2
# BAND             19.2    30.2   -2.7    7.3   11.0   218.5   141.3     9.9    18.5     5.3    31.7     4.7



"""
This file contains a class that holds info about a ForceField patch file.
It is often necessary to create multiple patch files and combine them into a single one,
to be pased as input to an AMSJob.
This class can do that.
"""

__all__ = ["ForceFieldPatch", "forcefield_params_from_kf"]


class ForceFieldPatch:
    """
    Class representing an Amber format force field patch file, as created by AMS
    """

    def __init__(self, text=None):
        """
        Creates an instance of the class ForceFieldPatch
        """
        # Set all instance variables
        self.comment = ""

        self.types = []
        self.bondtypes = []
        self.angletypes = []
        self.dihedraltypes = []
        self.impropertypes = []
        self.ljtypes = []

        self.typelines = []
        self.bondlines = []
        self.anglelines = []
        self.dihedrallines = []
        self.improperlines = []
        self.ljlines = []

        # Read the sections, and set the parameters
        if text is not None:
            lines = [l + "\n" for l in text.split("\n")]
            self.comment = lines[0]
            self._set_types(lines)
            self._set_bonds(lines)
            self._set_angles(lines)
            self._set_dihedrals(lines)
            self._set_impropers(lines)
            self._set_ljparams(lines)

    def get_text(self):
        """
        Create the full patch text
        """
        block = ""
        if len(self.types) == 0:
            return block
        block += self.comment
        block += "".join(self.typelines)
        block += "\n\n"
        block += "".join(self.bondlines)
        block += "\n"
        block += "".join(self.anglelines)
        block += "\n"
        block += "".join(self.dihedrallines)
        block += "\n"
        block += "".join(self.improperlines)
        block += "\n\n\n"
        block += "MOD4      RE\n"
        block += "".join(self.ljlines)
        block += "\nEND\n"
        return block

    def copy(self):
        """
        Returns a copy of self
        """
        ret = self.__class__()
        ret.comment = self.comment

        ret.types = self.types.copy()
        ret.bondtypes = self.bondtypes.copy()
        ret.angletypes = self.angletypes.copy()
        ret.dihedraltypes = self.dihedraltypes.copy()
        ret.impropertypes = self.impropertypes.copy()
        ret.ljtypes = self.ljtypes.copy()

        ret.typelines = self.typelines.copy()
        ret.bondlines = self.bondlines.copy()
        ret.anglelines = self.anglelines.copy()
        ret.dihedrallines = self.dihedrallines.copy()
        ret.improperlines = self.improperlines.copy()
        ret.ljlines = self.ljlines.copy()

        return ret

    def clear(self):
        """
        Empty self
        """
        self.comment = ""
        self._set_types([])
        self._set_bonds([])
        self._set_angles([])
        self._set_dihedrals([])
        self._set_impropers([])
        self._set_ljparams([])

    def __len__(self):
        """
        Returns the size of the patch
        """
        return len(self.types)

    def __str__(self):
        """
        Returns the patch as text
        """
        return self.get_text()

    def __add__(self, other):
        """
        Combine two patch files
        """
        ret = self.copy()
        if len(ret.comment) == 0:
            ret.comment = other.comment

        # Combine the lines from each section
        ret.typelines += [other.typelines[i] for i, t in enumerate(other.types) if not t in self.types]
        ret.bondlines += [other.bondlines[i] for i, b in enumerate(other.bondtypes) if not b in self.bondtypes]
        ret.anglelines += [other.anglelines[i] for i, t in enumerate(other.angletypes) if not t in self.angletypes]
        ret.dihedrallines += [
            other.dihedrallines[i] for i, t in enumerate(other.dihedraltypes) if not t in self.dihedraltypes
        ]
        ret.improperlines += [
            other.improperlines[i] for i, t in enumerate(other.impropertypes) if not t in self.impropertypes
        ]
        ret.ljlines += [other.ljlines[i] for i, t in enumerate(other.ljtypes) if not t in self.types]

        # Now adjust all the parameter info (not the lines)
        ret.bondtypes += [b for b in other.bondtypes if not b in self.bondtypes]
        ret.angletypes += [a for a in other.angletypes if not a in self.angletypes]
        ret.dihedraltypes += [d for d in other.dihedraltypes if not d in self.dihedraltypes]
        ret.impropertypes += [imp for imp in other.impropertypes if not imp in self.impropertypes]
        ret.ljtypes += [t for t in other.ljtypes if not t in self.types]
        ret.types += [t for t in other.types if not t in self.types]

        return ret

    def read_from_kf(self, kf):
        """
        Read patch infor from kf
        """
        if len(self) > 0:
            self.clear()
        npatches = kf.read("AMSResults", "Config.nPatches")
        patch = ForceFieldPatch()
        if npatches > 0:
            patchtext = kf.read("AMSResults", "Config.FFPatch(1)")
            patch += ForceFieldPatch(patchtext)

        for key in vars(patch):
            if "type" in key or "lines" in key or "comment" in key:
                self.__dict__[key] = patch.__dict__[key]

    def write_to_kf(self, kf):
        """
        Write the patch info to KF
        """
        kf.write("AMSResults", "Config.nPatches", 1)
        kf.write("AMSResults", "Config.FFPatch(1)", str(self))

    def _set_types(self, lines):
        """
        Read the atom types
        """
        types = []
        typelines = []
        for line in lines[1:]:
            words = line.split()
            if len(words) == 0:
                break
            types.append(words[0])
            typelines.append(line)
        self.types = types
        self.typelines = typelines

    def _set_bonds(self, lines):
        """
        Set the bond parameters from the list of lines
        """
        b, blines = self._read_atoms(lines, nats=2)
        self.bondtypes = b
        self.bondlines = blines

    def _set_angles(self, lines):
        """
        Set the angle parameters from list of lines
        """
        a, alines = self._read_atoms(lines, nats=3)
        self.angletypes = a
        self.anglelines = alines

    def _set_dihedrals(self, lines):
        """
        Set the dihedral parameters from list of lines
        """
        d, dlines = self._read_atoms(lines, nats=4)
        self.dihedraltypes = d
        self.dihedrallines = dlines

    def _set_impropers(self, lines):
        """
        Set improper parameters from list of lines
        """
        imp, implines = self._read_atoms(lines, nats=4, improper=True)
        self.impropertypes = imp
        self.improperlines = implines

    def _set_ljparams(self, lines):
        """
        Set the Lennard-Jones paramters from list of lines
        """
        lj, ljlines = self._read_LJtypes(lines)
        self.ljtypes = lj
        self.ljlines = ljlines

    @staticmethod
    def _read_atoms(lines, nats=2, improper=False):
        """
        Read the atoms from the patch lines for bond, angle, dihedral parameters

        * ``nats`` - Integer: 2 for bond info, 3 for angl info, 4 for dihedral info
        """
        step = 3
        positions = [2 + (i * step) for i in range(nats)]
        intervals = [((i * step), (i * step) + step - 1) for i in range(nats)]

        atomlist = []
        line_list = []
        for line in lines[1:]:
            if len(line) - 1 < max(positions):
                continue
            symbols = [line[i] for i in positions]
            if not False in [v == "-" for v in symbols[:-1]] and symbols[-1] == " ":
                # Disregard impropers
                found_improper = nats == 4 and line[14] == " "
                if not improper:
                    if found_improper:
                        continue
                else:
                    if not found_improper:
                        continue
                # Extract the atoms, and sort them if they are not improper angle atoms
                atoms = [line[t[0] : t[1]] for t in intervals]
                if not improper:
                    if sorted([atoms[0], atoms[-1]]) != [atoms[0], atoms[-1]]:
                        atoms = atoms[::-1]
                atomlist.append(atoms)
                line_list.append(line)
        return atomlist, line_list

    @staticmethod
    def _read_LJtypes(lines):
        """
        Read the LJ info from the patch file lines
        """
        start = False
        ljtypes = []
        ljlines = []
        for line in lines[1:]:
            if "MOD4      RE" in line:
                start = True
                continue
            if "END" in line:
                start = False
            if start:
                words = line.split()
                if len(words) == 0:
                    continue
                ljtypes.append(words[0])
                ljlines.append(line)
        return ljtypes, ljlines


def forcefield_params_from_kf(kf):
    """
    Read the parameters from kf
    """
    charges = kf.read("AMSResults", "Charges")
    alltypes = kf.read("AMSResults", "AtomTyping.atomTypes").split("\x00")
    indices = kf.read("AMSResults", "AtomTyping.atomIndexToType")
    types = [alltypes[i - 1] for i in indices]

    # Read the force field patch
    patch = ForceFieldPatch()
    patch.read_from_kf(kf)
    if len(patch) == 0:
        patch = None
    return charges, types, patch



import os
from os.path import join as opj

from scm.plams.core.basejob import SingleJob
from scm.plams.core.private import saferun
from scm.plams.core.results import Results
from scm.plams.core.settings import Settings

__all__ = ["DiracJob", "DiracResults"]


class DiracResults(Results):
    """A class for result of computation done with DIRAC."""

    _rename_map = {
        "DFCOEF": "$JN.dfcoef",
        "GRIDOUT": "$JN.grid",
        "dirac.xml": "$JN.xml",
    }

    def collect(self):
        """After collecting the files produced by job execution with parent method :meth:`Results.collect<scm.plams.core.results.Results.collect>` append the ``pam`` output to the regular output file."""
        Results.collect(self)
        pamfile = self.job._filename("out")
        process = saferun(["grep", "output file", pamfile], cwd=self.job.path)
        output = process.stdout.decode()
        diracfile = output.split(":")[-1].strip()
        if diracfile in self.files:
            pampath = opj(self.job.path, pamfile)
            diracpath = opj(self.job.path, diracfile)
            with open(pampath, "r") as f:
                pamoutput = f.readlines()
            with open(diracpath, "r") as f:
                diracoutput = f.readlines()
            with open(pampath, "w") as f:
                f.writelines(diracoutput)
                f.write("\n\n   " + "*" * 74 + "\n")
                f.write("   " + "*" * 30 + "  pam output  " + "*" * 30 + "\n")
                f.write("   " + "*" * 74 + "\n\n")
                f.writelines(pamoutput)
            os.remove(diracpath)
        self.refresh()


class DiracJob(SingleJob):
    """A class representing a single computational job with DIRAC."""

    _result_type = DiracResults
    _top = ["dirac"]
    _filenames = {"inp": "$JN.inp", "run": "$JN.run", "out": "$JN.out", "err": "$JN.err"}

    def __init__(self, **kwargs):
        SingleJob.__init__(self, **kwargs)
        self.settings.runscript.pam.noarch = True
        self.settings.runscript.pam.get = ["DFCOEF", "GRIDOUT", "dirac.xml"]

    def _get_ready(self):
        """Before generating runscript and input with parent method :meth:`SingleJob._get_ready<scm.plams.core.basejob.SingleJob._get_ready>` add proper ``mol`` and ``inp`` entries to ``self.settings.runscript.pam``. If already present there, ``mol`` will not be added."""
        s = self.settings.runscript.pam
        if "mol" not in s:
            s.mol = self.name + ".xyz"
            with open(opj(self.path, self.name + ".xyz"), "w") as f:
                f.write(str(len(self.molecule)) + "\n\n")
                for atom in self.molecule:
                    suffix = "b={block}" if hasattr(atom, "block") else ""
                    f.write(atom.str(suffix=suffix) + "\n")
        s.inp = self._filename("inp")
        SingleJob._get_ready(self)

    def get_input(self):
        """Transform all contents of ``input`` branch of ``settings`` into string with blocks, subblocks, keys and values.

        On the highest level alphabetic order of iteration is modified: keys occuring in class attribute ``_top`` are printed first. See :ref:`dirac-input` for details.
        """
        is_empty = lambda x: isinstance(x, Settings) and len(x) == 0

        def parse_key(key, value):
            ret = "." + key.upper() + "\n"
            if not (value is True or is_empty(value)):
                if isinstance(value, list):
                    for i in value:
                        ret += str(i) + "\n"
                else:
                    ret += str(value) + "\n"
            return ret

        def parse_block(block):
            enabler = "_en"
            ret = "**" + block.upper() + "\n"
            s = self.settings.input[block]
            for k, v in s.items():
                if not isinstance(v, Settings) or is_empty(v):
                    ret += parse_key(k, v)
            for k, v in s.items():
                if isinstance(v, Settings) and enabler in v:
                    ret += parse_key(k, v[enabler])
            for k, v in s.items():
                if isinstance(v, Settings) and len(v) > 0:
                    ret += "*" + k.upper() + "\n"
                    for kk, vv in v.items():
                        if kk != enabler:
                            ret += parse_key(kk, vv)
            return ret

        inp = ""
        for block in self._top:
            if block in self.settings.input:
                inp += parse_block(block)
        for block in self.settings.input:
            if block not in self._top:
                inp += parse_block(block)
        inp += "*END OF INPUT\n"
        return inp

    def get_runscript(self):
        """Generate a runscript. Returned string is a ``pam`` call followed by option flags generated based on ``self.settings.runscript.pam`` contents. See :ref:`dirac-runscript` for details."""
        r = self.settings.runscript.pam
        ret = "pam"
        for k, v in r.items():
            ret += " --%s" % k
            if v is not True:
                if isinstance(v, list):
                    ret += '="%s"' % " ".join(v)
                else:
                    ret += "=" + str(v)

        if self.settings.runscript.stdout_redirect:
            ret += " >" + self._filename("out")
        ret += "\n\n"
        return ret

    def check(self):
        """Check if the calculation was successful by examining the last line of ``pam`` output."""
        s = self.results.grep_output("exit           :")[0]
        status = s.split(":")[-1].strip()
        return status == "normal"



#!/usr/bin/env python

from scm.plams.core.settings import Settings
from scm.plams.mol.atom import Atom
from scm.plams.mol.molecule import Bond, Molecule
from scm.plams.tools.periodic_table import PT
from scm.plams.trajectories.sdffile import SDFTrajectoryFile, get_molecule

__all__ = ["SDFHistoryFile"]


class SDFHistoryFile(SDFTrajectoryFile):
    """
    Class representing an SDF file containing a molecular simulation history with varying numbers of atoms

    An instance of this class has the following attributes:

    *   ``file_object`` -- A Python :py:class:`file` object, referring to the actual SDF file
    *   ``position``    -- The frame to which the cursor is currently pointing in the SDF file
    *   ``mode``        -- Designates whether the file is in read or write mode ('r' or 'w')
    *   ``elements``    -- The elements of the atoms in the system at the current frame

    An |SDFHistoryFile| object behaves very similar to a regular file object.
    It has read and write methods (:meth:`read_next` and :meth:`write_next`)
    that read and write from/to the position of the cursor in the ``file_object`` attribute.
    If the file is in read mode, an additional method :meth:`read_frame` can be used that moves
    the cursor to any frame in the file and reads from there.
    The amount of information stored in memory is kept to a minimum, as only information from the current frame
    is ever stored.

    Reading and writing to and from the files can be done as follows::

        >>> from scm.plams import SDFHistoryFile

        >>> sdf = SDFHistoryFile('old.sdf')
        >>> mol = sdf.get_plamsmol()

        >>> sdfout = SDFHistoryFile('new.sdf',mode='w')

        >>> for i in range(sdf.get_length()) :
        >>>     crd,cell = sdf.read_frame(i,molecule=mol)
        >>>     sdfout.write_next(molecule=mol)

    The above script reads information from the SDF file ``old.sdf`` into the |Molecule| object ``mol``
    in a step-by-step manner.
    The |Molecule| object is then passed to the :meth:`write_next` method of the new |SDFHistoryFile|
    object corresponding to the new sdf file ``new.sdf``.

    The exact same result can also be achieved by iterating over the instance as a callable

        >>> sdf = SDFHistoryFile('old.sdf')
        >>> mol = sdf.get_plamsmol()

        >>> sdfout = SDFHistoryFile('new.sdf',mode='w')

        >>> for crd,cell in sdf(mol) :
        >>>     sdfout.write_next(molecule=mol)

    This procedure requires all coordinate information to be passed to and from the |Molecule| object
    for each frame, which can be time-consuming.
    It is therefore also possible to bypass the |Molecule| object when reading through the frames::

        >>> sdf = SDFHistoryFile('old.sdf')

        >>> sdfout = SDFHistoryFile('new.sdf',mode='w')

        >>> for crd,cell in sdf :
        >>>     sdfout.write_next(coords=crd,elements=sdf.elements)

    By default the write mode will create a minimal version of the SDF file, containing only elements
    and coordinates.
    Additional information can be written to the file by supplying additional arguments
    to the :meth:`write_next` method.
    The additional keywords `step` and `energy` trigger the writing of a remark containing
    the molecule name, the step number, the energy, and the lattice vectors.

        >>> mol = Molecule('singleframe.sdf')

        >>> sdfout = SDFHistoryFile('new.sdf',mode='w')
        >>> sdfout.set_name('MyMol')

        >>> sdfout.write_next(molecule=mol, step=0, energy=5.)
    """

    def __init__(self, filename, mode="r", fileobject=None, ntap=None):
        """
        Initiates an SDFHistoryFile object

        * ``filename``   -- The path to the SDF file
        * ``mode``       -- The mode in which to open the SDF file ('r' or 'w')
        * ``fileobject`` -- Optionally, a file object can be passed instead (filename needs to be set to None)
        * ``ntap``       -- If the file is in write mode, the number of atoms can be passed here
        """
        SDFTrajectoryFile.__init__(self, filename, mode, fileobject, ntap)

        self.input_elements = self.elements[:]

    def _read_coordinates(self, molecule):
        """
        Read the coordinates from file, and place them in the molecule
        """
        # Read lines until the end
        lines = []
        while True:
            line = self.file_object.readline()
            if len(line) == 0:
                return None, None  # End of file is reached
            lines.append(line)
            if len(line) < 4:
                continue
            if line[:4] == "$$$$":
                break

        # Get the mol part
        mol, restlines = get_molecule(lines)
        elements = [at.symbol for at in mol.atoms]

        # Get the coordinates and cell
        cell = mol.lattice
        if len(cell) == 0:
            cell = None
        if len(mol.bonds) > 0:
            conect = {}
            for bond in mol.bonds:
                iat = min(mol.index(bond))
                jat = max(mol.index(bond))
                if not iat in conect:
                    conect[iat] = []
                conect[iat].append(jat)
            self.conect = conect

        # If the elements changed, update the molecule
        if elements != self.elements:
            self.elements = elements
            self.coords = mol.as_array()
            # Rebuild the molecule (bonds will disappear for now)
            if isinstance(molecule, Molecule):
                for at in reversed(molecule.atoms):
                    molecule.delete_atom(at)
                molecule.properties = Settings()
                for el in elements:
                    atom = Atom(PT.get_atomic_number(el))
                    molecule.add_atom(atom)
        else:
            self.coords[:, :] = mol.as_array()

        # Read the additional data
        if self.include_historydata:
            historydata = {}
            # First find all entries (entries can run over multiple lines)
            entries = [i for i, line in enumerate(restlines[:-1]) if line[:4] == ">  <"] + [len(restlines) - 1]
            for i, iline in enumerate(entries[:-1]):
                key = restlines[iline].split("<")[1].split(">")[0]
                value = "".join(restlines[iline + 1 : entries[i + 1] - 1])
                value = value.strip()
                # Try to turn this into a float or integer?
                if value.isdigit():
                    value = int(value)
                else:
                    try:
                        value = float(value)
                    except ValueError:
                        pass
                historydata[key] = value
            self.historydata = historydata

        if isinstance(molecule, Molecule):
            self._set_plamsmol(self.coords, cell, molecule)

        return self.coords, cell

    def write_next(self, coords=None, molecule=None, cell=[0.0, 0.0, 0.0], conect=None, historydata=None):
        """
        Write frame to next position in trajectory file

        * ``coords``   -- A list or numpy array of (``ntap``,3) containing the system coordinates
        * ``molecule`` -- A molecule object to read the molecular data from
        * ``cell``     -- A set of lattice vectors or cell diameters
        * ``conect``   -- A dictionary containing connectivity info (not used)
        * ``historydata`` -- A dictionary containing additional variables to be written to the comment line

        The ``historydata`` dictionary can contain for example:
        ('Step','Energy'), the frame number and the energy respectively

        .. note::

                Either ``coords`` or ``molecule`` are mandatory arguments
        """
        if isinstance(molecule, Molecule):
            coords, cell, elements = self._read_plamsmol(molecule)[:3]
            self.elements = elements
        cell = self._convert_cell(cell)

        if not isinstance(molecule, Molecule):
            # Create the molecule?
            molecule = Molecule()
            for el, crd in zip(self.elements, coords):
                atom = Atom(symbol=el, coords=crd)
                molecule.add_atom(atom)
            if cell is not None:
                molecule.lattice = cell.tolist()
            # Add the bonds
            bondlist = []
            if conect is not None:
                for iat, neighbors in conect.items():
                    for jat in neighbors:
                        indices = tuple(sorted([iat, jat]))
                        if not indices in bondlist:
                            bondlist.append(indices)
                            bond = Bond(molecule.atoms[iat], molecule.atoms[jat])
                            molecule.add_bond(bond)

        self._write_moldata(molecule, historydata)

        self.position += 1



try:
    import dill as pickle  # noqa: F401
except ImportError:
    pass

from scm.plams import AMSJob, Settings
from scm.plams.unit_tests.test_helpers import skip_if_no_ams_installation


def test_hybrid_engine_input():
    """test :meth:`AMSJob.get_input` for writing sub engines in a hybrid engine block of ams."""
    AMSinput = """system
  Atoms
              C      -1.6447506665       1.4391568332       0.0000000000 
              C      -0.5773632247       0.3290989412      -0.0074515501 
              H      -1.2586274675       2.3064141938       0.5149500419 
              H      -2.5299121499       1.0840525749       0.5067446240 
              H      -1.8942698087       1.7054858887      -1.0164689035 
              H       0.3077982588       0.6842031995      -0.5141961741 
              H      -0.9634864236      -0.5381584194      -0.5224015919 
              H      -0.3278440824       0.0627698856       1.0090173534 
  End
  BondOrders
     2 1 1.0
     3 1 1.0
     4 1 1.0
     5 1 1.0
     6 2 1.0
     7 2 1.0
     8 2 1.0
  End
End

task SinglePoint

Engine hybrid
  energy
    term
      engineid ADF-1
      factor 0.5
      region *
    End
    term
      engineid DFTB-1
      factor 0.5
      region *
    End
  End
  engine ADF ADF-1
  EndEngine
  engine DFTB DFTB-1
  EndEngine

EndEngine

"""
    skip_if_no_ams_installation()
    job = AMSJob.from_input(AMSinput)
    assert job.get_input() == AMSinput


def test_list_block_input():
    """test :meth:`AMSJob.get_input` for writing list-blocks"""
    AMSinput = """system
  Atoms
              C      -1.6447506665       1.4391568332       0.0000000000 
  End
End

task SinglePoint

Engine hybrid
  energy
    term
      engineid ADF-1
      factor 0.5
      region *
    End
    term
      engineid ADF-2
      factor 0.5
      region *
    End
    term
      engineid DFTB-1
      factor 0.5
      region *
    End
  End
  engine ADF ADF-1
     Save TAPE10
     Save TAPE11
     Basis
     PerAtomType
     Symbol H
     End
     PerAtomType
     Symbol O
     End
     End
  EndEngine
  engine ADF ADF-2
  EndEngine
  engine DFTB DFTB-1
  EndEngine

EndEngine

"""
    skip_if_no_ams_installation()
    job = AMSJob.from_input(AMSinput)
    assert job.get_input() == AMSinput


def test_settings_to_molecule():
    s = Settings()
    s.input.ams.system.Atoms._1 = "         H       0.0000000000       0.0000000000       0.0000000000 "
    s.input.ams.system.Atoms._2 = "         O       1.0000000000       0.0000000000       0.0000000000 "

    t = s.copy()
    mol = AMSJob.settings_to_mol(t)[""]

    assert AMSJob(settings=s).get_input() == AMSJob(settings=t, molecule=mol).get_input()

    s = Settings()
    s.input.ams.system.Atoms._1 = [
        "         H       0.0000000000       0.0000000000       0.0000000000 ",
        "         O       1.0000000000       0.0000000000       0.0000000000 ",
    ]

    t = s.copy()
    mol = AMSJob.settings_to_mol(t)[""]

    assert AMSJob(settings=s).get_input() == AMSJob(settings=t, molecule=mol).get_input()

    # test for headers:
    s.input.ams.system.Atoms._h = "[A]"
    s.input.ams.system.Atoms._1 = [
        "         O      -0.0509000000      -0.2754000000       0.6371000000 ForceField.Charge=-0.834 ForceField.Type=OW",
        "         H       0.0157000000       0.5063000000       0.0531000000 ForceField.Charge=0.417 ForceField.Type=HW",
        "         H      -0.0055000000      -1.0411000000       0.0658000000 ForceField.Charge=0.417 ForceField.Type=HW",
        "         O       0.0981000000       1.7960000000      -1.2550000000 ForceField.Charge=-0.834 ForceField.Type=OW",
        "         H      -0.6686000000       2.2908000000      -1.5343000000 ForceField.Charge=0.417 ForceField.Type=HW",
        "         H       0.8128000000       2.3488000000      -1.5619000000 ForceField.Charge=0.417 ForceField.Type=HW",
    ]
    t = s.copy()
    # currently plams completely ignores the [A] unit specifier, it might also not print it if it does
    # get implemented and instead just convert the values in the molecule.
    del s.input.ams.system.Atoms._h

    mol = AMSJob.settings_to_mol(t)[""]
    assert AMSJob(settings=s).get_input() == AMSJob(settings=t, molecule=mol).get_input()



import sys
import os
import numpy
from scm.plams import AMSJob, Settings, Units, AMSAnalysisJob
from scm.plams import to_rdmol, init, finish


def main(filename, chargelines):
    """
    The main body of the script
    """
    # Get the molecular system
    job = AMSJob.load_external(filename)
    mol = job.molecule
    elements = [at.symbol for at in mol.atoms]

    # Read the temperature from KF
    T = job.results.readrkf("MDResults", "MeanTemperature")
    kBT = Units.constants["Boltzmann"] * T
    print("Average temperaturs %f K" % (T))

    # Read the charges from the input file
    ions = {}
    ioncharges = {}
    nions = {}
    formulas = {}
    for line in chargelines:
        words = line.split()
        if len(words) == 0:
            continue
        q = float(line.split(":")[0])
        atoms = [int(iat) for iat in line.split(":")[-1].split()]
        submol = mol.get_fragment(atoms)
        label = submol.label()
        if not label in ions.keys():
            ions[label] = []
            ioncharges[label] = q * Units.constants["electron_charge"]
            nions[label] = 0
            formulas[label] = submol.separate()[0].get_formula()
        ions[label] += atoms
        nions[label] += len(atoms)

    # Compute diffusion coefficient for each ion
    init()
    diffusion_coeffs = {}
    for label, atoms in ions.items():
        s = Settings()
        s.input.Task = "MeanSquareDisplacement"
        s.input.TrajectoryInfo.Trajectory.KFFilename = filename
        atsettings = [iat + 1 for iat in atoms]
        s.input.MeanSquareDisplacement.Atoms.Atom = atsettings

        job = AMSAnalysisJob(settings=s)
        results = job.run()
        D = results._kf.read("Slope(1)", "Final")
        diffusion_coeffs[label] = D
        units = results._kf.read("Slope(1)", "Final(units)")
        print(formulas[label], D, units)
    finish()

    # Compute the number density for each ion
    rho = {}
    for label, ni in nions.items():
        rho[label] = ni / mol.unit_cell_volume(unit="m")

    # Compute the ionic conductivity
    sigma = 0.0
    for label, D in diffusion_coeffs.items():
        s = ioncharges[label] ** 2 * rho[label] * D / kBT
        sigma += s
    return sigma


if __name__ == "__main__":
    if len(sys.argv) == 1 or sys.stdin.isatty():
        print("Usage: amspython get_ionic_conductivity.py path/to/ams.rkf < charges.in")
        sys.exit(0)
    chargelines = sys.stdin.readlines()
    filename = sys.argv[1]
    sigma = main(filename, chargelines)
    print("Ionic conductivity: %20.10e Siemens/m" % (sigma))



""" Deprecated, do not use """

from scm.plams import Settings


### ==== DFT SETTINGS ==== ###
def DFT():
    s = Settings()
    s.input.ams.task = "GeometryOptimization"
    s.input.adf.basis.type = "TZ2P"
    s.input.adf.basis.core = "None"
    s.input.adf.xc.hybrid = "B3LYP"
    s.input.adf.xc.Dispersion = "GRIMME3 BJDAMP"
    s.input.adf.Relativity.Level = "None"
    s.input.adf.NumericalQuality = "Good"
    s.input.adf.Symmetry = "NOSYM"
    s.input.ams.UseSymmetry = "No"
    s.input.adf.Unrestricted = "No"
    s.input.adf.SpinPolarization = 0
    s.input.ams.System.Charge = 0
    return s


### ==== DFTB SETTINGS ==== ###
def DFTB():
    s = Settings()
    s.input.ams.task = "GeometryOptimization"
    s.input.DFTB
    s.input.DFTB.Model = "GFN1-xTB"
    s.input.ams.System.Charge = 0
    return s


### ==== FREQ SETTINGS ==== ###
def frequencies():
    s = Settings()
    s.input.ams.properties.NormalModes = "Yes"
    s.input.ams.Properties.PESPointCharacter = "No"
    s.input.ams.NormalModes.ReScanFreqRange = "-1000 0"
    s.input.ams.PESPointCharacter.NegativeEigenvalueTolerance = -0.001
    return s
